import pytest
import itertools

from cosapp.patterns.visitor import Visitor, send
from cosapp.systems import System
from cosapp.drivers import RunOnce, NonLinearSolver, ValidityCheck
from cosapp.utils.testing import get_args, DummySystemFactory
from cosapp.tests.library.ports import XPort, X3Port


class NameCollector(Visitor):
    """Visitor collecting entity names"""
    def __init__(self):
        self.data = dict(
            (key, list())
            for key in ('systems', 'ports', 'drivers')
        )
    
    def visit_port(self, port) -> None:
        self.data['ports'].append(port.name)
    
    def visit_system(self, system) -> None:
        self.data['systems'].append(system.name)
    
    def visit_driver(self, driver) -> None:
        self.data['drivers'].append(driver.name)


class DataCollector(Visitor):
    """Visitor collecting miscellaneous data"""
    def __init__(self):
        self.data = dict(
            (key, dict())
            for key in ('systems', 'ports', 'drivers')
        )
    
    def visit_port(self, port) -> None:
        key = port.contextual_name
        self.data['ports'][key] = len(port)
    
    def visit_system(self, system) -> None:
        key = system.full_name()
        self.data['systems'][key] = system.size
        ports = list(system.inputs.values()) + list(system.outputs.values())
        send(self, ports)
    
    def visit_driver(self, driver) -> None:
        key = type(driver).__name__
        attr = self.data['drivers']
        attr.setdefault(key, 0)
        attr[key] += 1


@pytest.fixture
def system_1():
    def add_children(system, names):
        prefix = system.name
        for name in names:
            system.add_child(System(f"{prefix}{name}"))

    a = System('a')
    add_children(a, list('abc'))
    add_children(a.aa, list('ab'))
    add_children(a.ac, list('abc'))

    return a


@pytest.mark.parametrize("downwards, expected", [
    (True,  ['a', 'aa', 'aaa', 'aab', 'ab', 'ac', 'aca', 'acb', 'acc']),
    (False, ['aaa', 'aab', 'aa', 'ab', 'aca', 'acb', 'acc', 'ac', 'a']),
])
def test_System_send_visitor(system_1, downwards, expected):
    a = system_1
    visitor = NameCollector()
    a.send_visitor(visitor, downwards)

    assert visitor.data['ports'] == []
    assert visitor.data['drivers'] == []
    assert visitor.data['systems'] == expected


def test_System_send_visitor_default(system_1):
    """Test default behaviour of `System.send_visitor`.
    """
    a = system_1
    visitor = NameCollector()
    a.send_visitor(visitor)

    names = [
        'aaa', 'aab', 'aa',
        'ab',
        'aca', 'acb', 'acc', 'ac',
        'a',
    ]

    assert visitor.data['ports'] == []
    assert visitor.data['drivers'] == []
    assert visitor.data['systems'] == names

    # Send data collector
    visitor = DataCollector()
    a.send_visitor(visitor)

    assert visitor.data['drivers'] == {}
    assert visitor.data['systems'] == {
        'a': 9,
        'a.aa': 3,
        'a.ab': 1,
        'a.ac': 4,
        'a.aa.aaa': 1,
        'a.aa.aab': 1,
        'a.ac.aca': 1,
        'a.ac.acb': 1,
        'a.ac.acc': 1,
    }
    port_data = {}
    port_data.update(
        (f"{name}.{p}wards", 0)
        for name, p in itertools.product(names, ('in', 'out'))
    )
    port_data.update(
        (f"{name}.modevars_{p}", 0)
        for name, p in itertools.product(names, ('in', 'out'))
    )
    assert visitor.data['ports'] == port_data


def test_visitor_send(system_1):
    """Same as `test_System_send_visitor`, using function `visitor.send`.
    """
    a = system_1
    visitor = NameCollector()
    send(visitor, a.tree())

    assert visitor.data['ports'] == []
    assert visitor.data['drivers'] == []
    assert visitor.data['systems'] == [
        'aaa', 'aab', 'aa',
        'ab',
        'aca', 'acb', 'acc', 'ac',
        'a',
    ]


def test_send_DataCollector():
    """Send `DataCollector` visitor on composite system with drivers"""
    A = DummySystemFactory("A",
        inwards = [get_args('x', 1.0), get_args('y', 0.5)],
        outwards = [get_args('z', 0.0)],
        inputs = get_args(X3Port, 'p_in'),
        outputs = get_args(XPort, 'q_out'),
    )
    B = DummySystemFactory("B",
        inputs = get_args(X3Port, 'u'),
        outputs = get_args(X3Port, 'v'),
    )

    top = System('top')
    top.add_child(A('a'))
    top.add_child(B('b'))
    top.a.add_child(B('b'))

    top.add_driver(ValidityCheck('check'))
    top.a.add_driver(NonLinearSolver('solver'))
    top.b.add_driver(RunOnce('runner'))
    top.b.add_driver(ValidityCheck('check'))
    
    visitor = DataCollector()
    top.send_visitor(visitor)

    assert visitor.data['systems'] == {
        'top': 4,
        'top.a': 2,
        'top.b': 1,
        'top.a.b': 1,
    }
    port_data = {
        'top.inwards': 0,
        'top.outwards': 0,
        'b.inwards': 0,
        'b.outwards': 0,
        'a.inwards': 2,
        'a.outwards': 1,
        'a.p_in': 3,
        'a.q_out': 1,
        'b.u': 3,
        'b.v': 3,
    }
    names = [s.name for s in top.tree()]
    port_data.update(
        (f"{name}.modevars_{p}", 0)
        for name, p in itertools.product(names, ('in', 'out'))
    )
    assert visitor.data['ports'] == port_data
    assert visitor.data['drivers'] == {}

    for system in top.tree():
        for driver in system.drivers.values():
            send(visitor, driver.tree())

    assert visitor.data['drivers'] == {
        'NonLinearSolver': 1,
        'ValidityCheck': 2,
        'RunOnce': 1,
    }



"""Regression test from tutorials"""
import pytest

from cosapp.tests.library.systems import Resistor, Node, Source, Ground
from cosapp.systems import System
from cosapp.drivers import NonLinearSolver, RunSingleCase
from cosapp.recorders import DataFrameRecorder


class Circuit(System):
    """Simple resistor circuit

    --->--+--[R2]--+
          |        |
        [R1]     [R3]
          |        |
          +--------+---/ Ground
    """
    def setup(self):
        ground = self.add_child(Ground('ground', V=0.0))
        R1 = self.add_child(Resistor('R1', R=1000))
        R2 = self.add_child(Resistor('R2', R=500))
        R3 = self.add_child(Resistor('R3', R=250))
        
        Node.make(self, 'n1',
            incoming = [],
            outgoing = [R1, R2],
            pulling = {'I_in0': 'source'},
        )
        Node.make(self, 'n2',
            incoming = [R2],
            outgoing = [R3],
        )
        self.connect(ground.V_out, R1.V_out)
        self.connect(ground.V_out, R3.V_out)
        
        self.exec_order = ('ground', 'n1', 'n2', 'R1', 'R2', 'R3')

        # design methods
        self.add_inward('V1_design', 0.0)
        self.add_inward('V2_design', 0.0)
        self.add_design_method('R1').add_unknown('R1.R').add_equation('n1.V == V1_design')
        self.add_design_method('R2').add_unknown('R2.R').add_equation('n2.V == V2_design')


class LegacyCircuit(System):
    """Legacy circuit model, as originally given in tutorials"""
    def setup(self):
        n1 = self.add_child(Node('n1', n_in=1, n_out=2), pulling={'I_in0': 'I_in'})
        n2 = self.add_child(Node('n2'))
        
        R1 = self.add_child(Resistor('R1', R=1000), pulling={'V_out': 'Vg'})
        R2 = self.add_child(Resistor('R2', R=500))
        R3 = self.add_child(Resistor('R3', R=250), pulling={'V_out': 'Vg'})
        
        self.connect(R1.V_in, n1.inwards, 'V')
        self.connect(R2.V_in, n1.inwards, 'V')
        self.connect(R1.I, n1.I_out0)
        self.connect(R2.I, n1.I_out1)
        
        self.connect(R2.V_out, n2.inwards, 'V')
        self.connect(R3.V_in, n2.inwards, 'V')
        self.connect(R2.I, n2.I_in0)
        self.connect(R3.I, n2.I_out0)
        
        self.exec_order = ('n1', 'n2', 'R1', 'R2', 'R3')

        # design methods
        self.add_inward('V1_design', 0.0)
        self.add_inward('V2_design', 0.0)
        self.add_design_method('R1').add_unknown('R1.R').add_equation('n1.V == V1_design')
        self.add_design_method('R2').add_unknown('R2.R').add_equation('n2.V == V2_design')


@pytest.fixture
def model():
    """Legacy model, as originally given in tutorials

     +----+--[R2]--+
     |    |        |
    (S)  [R1]     [R3]
     |    |        |
     +----+--------+---/ Ground
    """
    model = System('model')
    model.add_child(Source('source', I=0.1))
    model.add_child(Ground('ground', V=0.0))
    model.add_child(LegacyCircuit('circuit'))

    model.connect(model.source.I_out, model.circuit.I_in)
    model.connect(model.ground.V_out, model.circuit.Vg)

    model.circuit.R1.R = 1e3
    model.circuit.R2.R = 0.5e3
    model.circuit.R3.R = 0.25e3

    return model


@pytest.fixture
def circuit():
    circuit = Circuit('circuit')
    circuit.R1.R = 1e3
    circuit.R2.R = 0.5e3
    circuit.R3.R = 0.25e3
    circuit.source.I = 0.1
    return circuit


def test_LegacyCircuit_solve(model):
    model.add_driver(NonLinearSolver('solver'))
    model.run_drivers()

    assert model.source.I == 0.1
    assert model.circuit.R1.R == 1000
    assert model.circuit.R2.R == 500
    assert model.circuit.R3.R == 250
    assert model.circuit.R1.I.I == pytest.approx(0.04285714286)
    assert model.circuit.R2.I.I == pytest.approx(0.05714285714)
    assert model.circuit.R3.I.I == pytest.approx(0.05714285715)
    assert model.circuit.n1.V == pytest.approx(42.85714286)
    assert model.circuit.n2.V == pytest.approx(14.28571429)


def test_LegacyCircuit_design_multipoint(model):
    solver = model.add_driver(NonLinearSolver('solver', tol=1e-9))
    model.run_drivers()  # balance system

    # Add driver to set boundary conditions on point 1
    point1 = solver.add_child(RunSingleCase('point1'))
    point1.design.extend(model.circuit.design_methods['R2'])

    model.circuit.V2_design = 8.
    point1.set_values({
        'source.I': 0.08, 
        'ground.V': 0,
    })

    # Same as previous for a second point
    point2 = solver.add_child(RunSingleCase('point2')) 
    point2.design.extend(model.circuit.design_methods['R1'])

    model.circuit.V1_design = 50.
    point2.set_values({
        'source.I': 0.15,
        'ground.V': 0,
    })

    solver.add_recorder(
        DataFrameRecorder(includes=['*n?.V', '*R', 'source.I'], excludes='*R3*')
    )

    model.run_drivers()

    data = solver.recorder.export_data()
    assert data.at[0, 'circuit.n2.V'] == pytest.approx(8)
    assert data.at[1, 'circuit.n1.V'] == pytest.approx(50)
    assert model.circuit.R1.R == pytest.approx(5000 / 9)
    assert model.circuit.R2.R == pytest.approx(5250 / 9)


def test_Circuit_solve(circuit):
    circuit.add_driver(NonLinearSolver('solver'))
    circuit.run_drivers()

    assert circuit.source.I == 0.1
    assert circuit.R1.R == 1000
    assert circuit.R2.R == 500
    assert circuit.R3.R == 250
    assert circuit.R1.I.I == pytest.approx(0.04285714286)
    assert circuit.R2.I.I == pytest.approx(0.05714285714)
    assert circuit.R3.I.I == pytest.approx(0.05714285715)
    assert circuit.n1.V == pytest.approx(42.85714286)
    assert circuit.n2.V == pytest.approx(14.28571429)


def test_Circuit_design_multipoint(circuit):
    solver = circuit.add_driver(NonLinearSolver('solver', tol=1e-9))
    circuit.run_drivers()  # balance system

    # Add driver to set boundary conditions on point 1
    point1 = solver.add_child(RunSingleCase('point1'))
    point1.design.extend(circuit.design_methods['R2'])

    circuit.V2_design = 8.
    point1.set_values({
        'source.I': 0.08, 
        'ground.V': 0,
    })

    # Same as previous for a second point
    point2 = solver.add_child(RunSingleCase('point2')) 
    point2.design.extend(circuit.design_methods['R1'])

    circuit.V1_design = 50.
    point2.set_values({
        'source.I': 0.15,
        'ground.V': 0,
    })

    solver.add_recorder(
        DataFrameRecorder(includes=['n?.V', '*R', 'source.I'], excludes='*R3*')
    )

    circuit.run_drivers()

    data = solver.recorder.export_data()
    assert data.at[0, 'n2.V'] == pytest.approx(8)
    assert data.at[1, 'n1.V'] == pytest.approx(50)
    assert circuit.R1.R == pytest.approx(5000 / 9)
    assert circuit.R2.R == pytest.approx(5250 / 9)



"""Module containing base classes for user-defined classes.
"""

from cosapp.ports import Port, Scope
from cosapp.systems import System
from cosapp.drivers import Driver
from cosapp.ports.connectors import BaseConnector
from cosapp.utils.surrogate_models import SurrogateModel

# Custom exceptions
from cosapp.ports import ConnectorError, ScopeError, UnitError



from __future__ import annotations
import enum
from typing import Optional, Union, Dict, Tuple, List, Set, NamedTuple
from collections.abc import Collection

from cosapp.core.numerics.boundary import Unknown
from cosapp.core.numerics.basics import MathematicalProblem
from cosapp.utils.helpers import check_arg
from cosapp.utils.parsing import multi_split

import logging
logger = logging.getLogger(__name__)


class SystemAnalyzer:
    """Class containing data collected on a system,
    to be shared between different drivers.
    """
    __slots__ = ('__system',)

    def __init__(self, system: Optional["cosapp.systems.System"] = None):
        self.__reset()
        self.system = system

    @property
    def system(self) -> "cosapp.systems.System":
        "System: system of interest"
        return self.__system

    @system.setter
    def system(self, system) -> None:
        if system is self.__system:
            return
        from cosapp.base import System
        check_arg(system, 'system', (System, type(None)))
        self.__reset(system)

    def __reset(self, system=None) -> None:
        """Reset system and data, with no type check"""
        self.__system = system
        self.clear()

    def clear(self) -> None:
        """Hook function to clear internal data, if any.
        Called at object instanciation.
        """
        pass

    def check_system(self) -> None:
        """Check that object is associated to a system.
        
        Raises
        ------
        `ValueError` if object is not associated to a system.
        """
        if self.__system is None:
            raise ValueError("object is not associated to any system")


def dealias_problem(problem: MathematicalProblem, name=None) -> MathematicalProblem:
    """Resolve unknown aliasing in `problem` due to
    pulling connectors (if any) in context system.

    Parameters:
    -----------
    - problem [MathematicalProblem]:
        Mathematical problem to be de-aliased.
    - name [str, optional]:
        Name of output `MathematicalProblem` object.
    
    Returns:
    --------
    MathematicalProblem:
        New mathematical problem, with identical equations as source
        problem, but de-aliased unknowns.
    """
    if problem.context is None:
        raise ValueError(f"problem is not defined on a system")
    if not name:
        name = f"{problem.name}[filtered]"
    from cosapp.base import System
    context: System = problem.context
    input_mapping = context.input_mapping

    def get_free_unknown(unknown: Unknown) -> Union[Unknown, None]:
        """Checks if `unknown` is aliased by pulling.
        If so, returns alias unknown; else, returns original unknown.
        """
        if unknown.context is context:
            contextual_name = unknown.basename
        else:
            contextual_name = f"{context.get_path_to_child(unknown.context)}.{unknown.basename}"
        try:
            alias = input_mapping[contextual_name]
        except KeyError:
            logger.warning(f"Skip connected unknown {contextual_name!r}")
            return None
        aliased = (alias is not unknown.ref)
        if aliased:
            try:
                path = context.get_path_to_child(alias.context)
            except ValueError:
                # `alias.context` is not in `system` tree
                logger.warning(
                    f"Unknown {unknown.contextual_name()!r} is aliased by {alias.contextual_name!r}"
                    f", defined outside the context of {context.name!r}"
                    f"; it is likely to be overwritten after the computation."
                )
            else:
                alias_contextual_name = f"{path}.{alias.name}" if path else alias.name
                unknown = unknown.transfer(context, alias_contextual_name)
                logger.info(f"Replace unknown {contextual_name!r} by {alias_contextual_name!r}")

        return unknown
    
    # Create output
    filtered = context.new_problem(name)
    # Add equations
    filtered.extend(problem, unknowns=False)
    # Add unknowns
    for key, unknown in problem.unknowns.items():
        free_unknown = get_free_unknown(unknown)
        if free_unknown is None:
            continue
        aliased = (free_unknown is not unknown)
        if aliased:
            key = free_unknown.name
            filtered.unknowns[key] = free_unknown
        else:
            filtered.unknowns[key] = unknown.copy()

    return filtered


class DesignProblemHandler(SystemAnalyzer):
    """Class managing tied design and off-design problems,
    including unknown aliasing.
    """
    def clear(self) -> None:
        """Reset inner problems"""
        self.design = self.new_problem('design')
        self.offdesign = self.new_problem('offdesign')

    @classmethod
    def make(cls, design: MathematicalProblem, offdesign: MathematicalProblem) -> DesignProblemHandler:
        handler = cls(design.context)
        handler.problems = (design, offdesign)
        return handler

    @property
    def problems(self) -> Tuple[MathematicalProblem, MathematicalProblem]:
        """Tuple[MathematicalProblem, MathematicalProblem]: design and off-design problems as a tuple"""
        return self.design, self.offdesign

    @problems.setter
    def problems(self, problems: Tuple[MathematicalProblem, MathematicalProblem]) -> None:
        """Setter for (design, offdesign) tuple"""
        wrong_type = lambda problem: not isinstance(problem, MathematicalProblem)
        if any(map(wrong_type, problems)):
            raise TypeError(
                f"expected `MathematicalProblem` instances; got {tuple(map(type, problems))}."
            )
        self.check_context(*problems)
        self.design, self.offdesign = problems

    def check_context(self, design: MathematicalProblem, offdesign: MathematicalProblem) -> None:
        ok = design.context is offdesign.context is self.system
        if not ok:
            raise ValueError(
                f"Design and off-design problems must be defined in the context of {self.system}."
            )

    def new_problem(self, name: str) -> MathematicalProblem:
        """Create new `MathematicalProblem` instance"""
        return MathematicalProblem(name, self.system)

    def copy_problems(self, prune=True) -> Tuple[MathematicalProblem, MathematicalProblem]:
        """Export design and off-design problems.
        
        Parameters:
        -----------
        - prune, Optional[bool]:
            If `True` (default), resolve unknown aliasing first.
            If `False`, returned problems are copies of object attributes.

        Results:
        --------
        - (design, offdesign): Copies of design and off-design problems,
            as a tuple of `MathematicalProblem` objects.
        """
        if prune:
            design = dealias_problem(self.design, name='design')
            offdesign = dealias_problem(self.offdesign, name='offdesign')
        else:
            design = self.design.copy()
            offdesign = self.offdesign.copy()
        return design, offdesign

    def copy(self, prune=True) -> DesignProblemHandler:
        """Returns a copy of the current object.
        
        Parameters:
        -----------
        - prune, Optional[bool]:
            If `True` (default), resolve unknown aliasing first.
            If `False`, returned handler contains copies of object problems.
        """
        design, offdesign = self.copy_problems(prune)
        return DesignProblemHandler.make(design, offdesign)

    def prune(self) -> None:
        """Remove connected unknowns and resolve unknown aliasing
        in design and off-desing problems.
        """
        self.design, self.offdesign = self.copy_problems(prune=True)

    def merged_problem(self, name="merged", offdesign_prefix="offdesign", copy=True) -> MathematicalProblem:
        """Merge design and off-design problems into a single `MathematicalProblem` instance.

        Parameters
        ----------
        - name [str, optional]:
            Merged problem name (default: 'merged').
        - offdesign_prefix [str, optional]:
            If not empty or `None`, applies a prefix to dict keys in off-design unknowns and equations.
        - copy [bool, optional]:
            Perform copies if `True` (default).
        """
        design, offdesign = self.copy_problems() if copy else self.problems

        def check(attr, kind):
            design_attr = getattr(design, attr)
            offdesign_attr = getattr(offdesign, attr)
            common = set(design_attr).intersection(offdesign_attr)
            if common:
                if len(common) > 1:
                    names = ", ".join(repr(v) for v in sorted(common))
                    names = f"({names}) are"
                    kind += "s"
                else:
                    names = f"{common.pop()!r} is"
                raise ValueError(
                    f"{names} defined as design and off-design {kind}"
                )
        check('unknowns', kind='unknown')
        check('residues', kind='equation')
        
        merged = self.new_problem(name)
        no_rename = lambda name: name
        local_name = (lambda name: f"{offdesign_prefix}[{name}]") if offdesign_prefix else no_rename

        def add_problem(problem: MathematicalProblem, rename_unknowns=True) -> None:
            nonlocal merged
            rename = local_name if rename_unknowns else no_rename

            # Add unknowns
            for name in list(problem.unknowns.keys()):
                merged.unknowns[rename(name)] = problem.unknowns.get(name)
            
            # Add residues
            for name in list(problem.residues.keys()):
                merged.residues[local_name(name)] = problem.residues.get(name)

            for name, residue in problem.get_target_residues().items():
                merged.residues[local_name(name)] = residue

        add_problem(design, rename_unknowns=False)
        add_problem(offdesign, rename_unknowns=True)
        return merged

    def extend(self, other: DesignProblemHandler, prune=True, copy=True, overwrite=False) -> DesignProblemHandler:
        """Extend both design and off-design problems from `other` handler.
        
        Parameters
        ----------
        - prune [bool, optional]:
            If `True` (default), added problems are pruned before being added.
        - copy [bool, optional]:
            Determines whether problem copies should be made before extension.
        - overwrite [bool, optional]:
            Overwrite option, forwarded to `MathematicalProblem.extend`.
        """
        if prune or copy:
            design, offdesign = other.copy_problems(prune)
        else:
            design, offdesign = other.problems
        # Extend inner problems; copy is unnecessary at this point,
        # as `design` and `offdesign` are consistent with argument `copy`.
        options = dict(copy=False, overwrite=overwrite)
        self.design.extend(design, **options)
        self.offdesign.extend(offdesign, **options)
        return self


@enum.unique
class ConstraintType(enum.Enum):
    """Enum covering constraint types"""
    GE = {
        'operator': ">=",
        'sort': (lambda lhs, rhs: (lhs, rhs)),
    }
    LE = {
        'operator': "<=",
        'sort': (lambda lhs, rhs: (rhs, lhs)),
    }
    EQ = {
        'operator': "==",
        'sort': (lambda lhs, rhs: (lhs, rhs)),
    }

    def expression(self, lhs: str, rhs: str) -> str:
        return "{} - ({})".format(*self.sort(lhs, rhs))

    def sort(self, lhs: str, rhs: str) -> Tuple[str, str]:
        return self.value['sort'](lhs, rhs)

    def __str__(self) -> str:
        return self.description

    @property
    def operator(self) -> str:
        return self.value['operator']

    @property
    def description(self) -> str:
        return f"Constraint of the kind `lhs {self.operator} rhs`"

    @property
    def is_inequality(self) -> bool:
        return self.operator != "=="


class Constraint(NamedTuple):
    """Named tuple representing a non-negative constraint
    of the kind `lhs <op> rhs`, where `<op>` is either
    `==` (equality) or `>=` (inequality constraint),
    depending on Boolean attribute `is_inequality`.

    Attributes:
    -----------
    - lhs [str]: left-hand side.
    - rhs [str]: right-hand side.
    - is_inequality [bool].

    Properties:
    -----------
    - expression [str]: non-negative constraint `lhs - rhs`.
    """
    lhs: str
    rhs: str
    is_inequality: bool = True

    @property
    def expression(self) -> str:
        return f"{self.lhs} - ({self.rhs})"

    def __str__(self) -> str:
        op = ">=" if self.is_inequality else "=="
        return f"{self.expression} {op} 0"


class ConstraintParser:

    @classmethod
    def parse(cls, expression: Union[str, List[str]]) -> Set[Constraint]:
        """Parse a string expression or a list thereof as a
        set of non-negative constraints to be used in solvers.

        Parameters:
        -----------
        - expression [str or List[str]]:
            Human-readable equality or inequality constraints,
            such as 'x >= y', '0 < alpha < 1', or a list thereof.
        
        Returns:
        --------
        - constraints [Set[Constraint]]:
            Set of `Constraint` named tuple objects.
        """
        check_arg(expression, 'expression', (str, Collection))
        ctypes = cls.types()
        not_in_sides = list(ctypes) + ["="]
        
        def side_ok(side: str) -> bool:
            return not any(nogo in side for nogo in not_in_sides) and side.strip()

        def parse_single(expression: str) -> Set[Constraint]:
            check_arg(expression, 'expression', str)
            constraints = set()

            for operator in "<>":
                expression = expression.replace(f"{operator}=", operator)

            expressions, operators = multi_split(expression, ctypes.keys())

            for lhs, rhs, operator in zip(expressions, expressions[1:], operators):
                ok = side_ok(lhs) and side_ok(rhs)
                ctype = ctypes[operator]
                constraint = Constraint(
                    *ctype.sort(lhs, rhs),
                    ctype.is_inequality,
                )
                if not ok:
                    raise ValueError(f"Invalid constraint {constraint.expression}")
                constraints.add(constraint)

            return constraints
        
        expressions = [expression] if isinstance(expression, str) else expression

        constraints = set()
        for expression in expressions:
            constraints |= parse_single(expression)

        return constraints

    @classmethod
    def types(cls) -> Dict[str, ConstraintType]:
        return {
            ">" : ConstraintType.GE,
            "<" : ConstraintType.LE,
            "<=": ConstraintType.LE,
            "==": ConstraintType.EQ,
            ">=": ConstraintType.GE,
        }



"""Build a VISjs view of a System."""

import os
from typing import Dict, List, Optional, Tuple, Union

from cosapp.drivers import Driver
from cosapp.ports.port import BasePort, ExtensiblePort
from cosapp.systems import System
from cosapp.tools import templates

from .baseRenderer import BaseRenderer


class VisJsRenderer(BaseRenderer):
    """
    Utility class to export a system as HTML using vis.JS library.

    Parameters
    ----------
    system : System
        System to export
    embeddable: bool, optional
        Is the HTML to be embedded in an existing page? Default: False
    """

    __globals = None

    def __init__(self, system: System, embeddable=False):
        super().__init__(system, embeddable)

    @staticmethod
    def get_system_label(system: System):
        label = system.name
        lmax = 7
        if len(label) > lmax:
            dots = "\u2026"  # single character
            label = label[:lmax-1] + dots
        return label

    def get_system_title(self, system: System):
        head = self.system
        if system is head:
            name = head.name
        else:
            name = head.get_path_to_child(system)
        return f"{name} - {system.__class__.__name__}"

    def get_system_group(self, system: System):
        head = self.system
        if system is head:
            name = f".{head.name}"
        else:
            path = head.get_path_to_child(system.parent)
            name = f"{head.name}.{path}" if path else head.name
        return name

    def get_data(self) -> Dict:
        """Convert the `System` in a dictionary with 4 keys: title, nodes, edges and groups.

        The title is a string with the Module name.
        The nodes is a list of dictionaries defining node for vis JS library.
        The edges is a list of dictionaries defining edge for vis JS library.
        The groups is a list of string defining the nodes clusters for vis JS library.

        Parameters
        ----------
        focus_system : System or None
            System on which the visualization should focus by default

        Returns
        -------
        dict
            Dictionary containing elements to draw the `System` using visJS library
        """
        head = self.system

        def get_component2id(
            system: System,
            cmpt2id: Dict[Union[System, BasePort], int],
            node_id: int,
            edges: List[Dict],
        ) -> Tuple[Dict[Union[System, BasePort], int], int, List[Dict]]:
            """Set the mapping between the `Module` and their unique id.

            Parameters
            ----------
            system : Module
                The parent `Module` for which id must be set
            cmpt2id : Dict[Union[Module, BasePort], int]
                Mapping of `Module` or port and its id
            node_id : int
                Latest used id
            edges : List[Dict]
                List of connection edges

            Returns
            -------
            tuple (List[Dict], Dict[Union["Module", "BasePort"], int], int)
                The first is the mapping between each involved system or port and an unique id,
                the second is the latest used id and
                the third index is the list of edges (one per connector).
            """

            def get_driver2id(driver: Driver, cmpt2id: dict, node_id: int, edges: List[dict]):
                if driver not in cmpt2id:
                    cmpt2id[driver] = node_id
                    node_id += 1

                previous_driver = None
                for subdriver in driver.children.values():
                    cmpt2id, node_id, edges = get_driver2id(
                        subdriver, cmpt2id, node_id, edges
                    )
                    # Connection between same level drivers
                    if previous_driver is not None: 
                        edges.append(
                            {
                                "from": cmpt2id[previous_driver],
                                "to": cmpt2id[subdriver],
                                "arrows": "to",
                                "title": "",
                                "hidden": True,
                                "physics": False,
                            }
                        )
                    previous_driver = subdriver
                    # Connection with parent driver
                    edges.append(
                        {
                            "from": cmpt2id[driver],
                            "to": cmpt2id[subdriver],
                            "title": "",
                            "hidden": True,
                            "physics": False,
                        }
                    )

                return cmpt2id, node_id, edges

            def connectors_to_visJS(
                system: System,
                component_id: Optional[Dict[Union[System, BasePort], int]] = None,
                node_id: int = 1,
            ) -> Tuple[List[Dict], Dict[Union[System, BasePort], int], int]:
                """Build the dictionary the list of edges.

                Parameters
                ----------
                system : System
                    System owning the connectors
                component_id : Dict[Union[System, Port], int], optional
                    Mapping of `System` or `Port` and its unique id; default None
                node_id : int, optional
                    Latest used id; default is 1.

                Returns
                -------
                tuple (List[Dict], Dict[System | BasePort, int], int)
                    First element is the mapping between each involved system/port and a unique id.
                    second element is the latest id used;
                    Third element is the list of edges (one per connector);
                """
                edges = list()

                for connection in system.all_connectors():
                    supplier = connection.source.owner
                    if supplier.children:  # Insert port as node
                        supplier = connection.source

                    if supplier not in component_id:
                        component_id[supplier] = node_id
                        node_id += 1

                    target = connection.sink.owner
                    if target.children:  # Insert port as node
                        target = connection.sink

                    if target not in component_id:
                        component_id[target] = node_id
                        node_id += 1

                    edge = {
                        "from": component_id[supplier],
                        "to": component_id[target],
                        "arrows": "to",
                        "title": "",
                    }
                    if connection.is_mirror():
                        edge["title"] = type(connection.source).__name__
                    else:
                        to = "&#8594;"  # rightarrow
                        edge["title"] = ", ".join(
                            origin if origin == target else f"{origin}{to}{target}"
                            for target, origin in connection.mapping.items()
                        ).join("[]")

                    # if supplier or target is top-system, set edge.length = 0
                    if (
                        connection.source.owner is connection.sink.owner.parent
                        or connection.source.owner.parent is connection.sink.owner
                    ):
                        edge["length"] = 0

                    edges.append(edge)

                return edges, component_id, node_id

            for driver in system.drivers.values():
                cmpt2id, node_id, edges = get_driver2id(driver, cmpt2id, node_id, edges)

            for child in system.children.values():
                if child not in cmpt2id:
                    cmpt2id[child] = node_id
                    node_id += 1

                if child.children:
                    cmpt2id, node_id, edges = get_component2id(child, cmpt2id, node_id, edges)

            local_edges, cmpt2id, node_id = connectors_to_visJS(system, cmpt2id, node_id)
            edges.extend(local_edges)

            return cmpt2id, node_id, edges

        cmpt2id = dict()
        id = 1
        edges = list()
        cmpt2id, id_init, edges = get_component2id(head, cmpt2id, id, edges)

        nodes = list()
        # Add hidden node for the top system
        nodes.append(
            dict(
                name=head.name,
                label=self.get_system_label(head),
                title=self.get_system_title(head),
                id=id_init,
                level=0,
                hidden=True,
                physics=False,
                group=f".{head.name}",
            )
        )

        # Add edges from uppest level to its children but deactivate physics on
        # them. The goal is to have them available for the hierarchical view but
        # not playing any role in the default layout.
        for child in head.children.values():
            edges.append(
                {
                    "from": id_init,
                    "to": cmpt2id[child],
                    "hidden": True,
                    "physics": False,
                }
            )

        groups = list()

        for c, id in cmpt2id.items():

            if isinstance(c, BasePort):
                ref = dict(
                    title=f"{c.full_name(trim_root=True)}",
                    id=id,
                    group=head.get_path_to_child(c.owner, trim_top=False),
                    mass=2,
                )
                if not isinstance(c, ExtensiblePort):
                    ref["title"] += f" - {type(c).__name__}"

                # Add edge to bind to central owner node
                edge = {
                    "from": id,
                    "to": cmpt2id[c.owner] if c.owner in cmpt2id else id_init,
                    "hidden": True,
                }
                if id_init in (edge["from"], edge["to"]):
                    edge["physics"] = False
                edges.append(edge)

            elif isinstance(c, Driver):
                # Add driver node
                ref = dict(
                    label=type(c).__name__,
                    title=f"{c.name} - {type(c).__name__}",
                    id=id,
                    level=-0.7,
                    shape="box",
                    hidden=True,
                    physics=False,
                    group=head.get_path_to_child(c.owner, trim_top=False),
                )

                # Add edge to Module owner if top driver
                if c.parent is None:
                    edges.append(
                        {
                            "from": cmpt2id[c.owner] if c.owner in cmpt2id else id_init,
                            "to": cmpt2id[c],
                            "dashes": True,
                            "hidden": True,
                            "physics": False,
                        }
                    )
                else:
                    driver_depth = len(list(c.path_to_root())) - 1
                    ref["level"] += 0.2 * driver_depth

            else:
                # c is a System
                ref = dict(
                    name=c.name,
                    label=self.get_system_label(c),
                    title=self.get_system_title(c),
                    id=id,
                    widthConstraint={'maximum': '10px'},
                )

                if c.children:  # Component containing component => Cluster
                    ref["hidden"] = True

                    # Add edge to force binding to central system node
                    for child in c.children.values():
                        edges.append({"from": id, "to": cmpt2id[child], "hidden": True})

                ref["group"] = self.get_system_group(c)

            group = ref["group"]
            ref.setdefault("level", 0)
            ref["level"] += group.count(".") + 1  # Set hierarchical level

            nodes.append(ref)
            if group not in groups:
                groups.append(group)

        # Sort the groups by system to ensure that levels are clustered bottom-up
        groups.sort(reverse=True)

        data = {
            "title": self.get_system_label(head),
            "nodes": nodes,
            "edges": edges,
            "groups": groups,
        }

        return data

    def html_content(self) -> str:
        """Returns vis.JS HTML content of renderer"s system as a character string."""
        elements = self.get_data()
        common = self.get_globals()
        elements["visJS"] = common["visJS"]
        elements["visCSS"] = common["visCSS"]
        html_begin_tags = common["html_begin_tags"]
        html_end_tags = common["html_end_tags"]
        template = common["template"]
        
        content = template.render(elements)
        if not self.embeddable:
            html_begin_tags = html_begin_tags.replace("{{title}}", elements["title"])
            content = html_begin_tags + content + html_end_tags

        return content

    @classmethod
    def html_resources(cls) -> Dict[str, str]:
        """
        Return a dictionary of two keys.

        - `visJS` key holds the content of vis.js library.
        - `visCSS` key holds the css style of plot.

        Returns
        -------
        Dict[str, Dict]
            The necessary resources to render Jinja template
        """
        ressource_dir = os.path.dirname(os.path.abspath(templates.__file__))
        libs_dir = os.path.join(ressource_dir, "lib")

        with open(os.path.join(libs_dir, "vis.min.js"), "rb") as f:
            vis_js = f.read().decode("utf-8")
        with open(os.path.join(libs_dir, "vis.min.css"), "rb") as f:
            vis_css = f.read().decode("utf-8")

        return {"visJS": vis_js, "visCSS": vis_css}
    
    @classmethod
    def html_template(cls):
        """
        Return the Jinja template used to create HTML file.

        Returns
        -------
        Template
        """
        try:
            from jinja2 import Environment, PackageLoader
        except ImportError:
            raise ImportError("jinja2 needs to be installed to export to vis JS.")

        env = Environment(
            loader=PackageLoader("cosapp.tools", "templates"),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        template = env.get_template("system_repr.html")
        return template
    
    @classmethod
    def get_globals(cls):
        cg = cls.__globals
        if cg is None:
            cg = super().get_globals()
        return cg


def visjs_html(system: System, embeddable=False) -> str:
    """Returns vis.JS HTML content of `system`.

    Parameters
    ----------
    system : System
        System to export

    embeddable: bool, optional
        Is the HTML to be embedded in an existing page? Default: False
    """
    renderer = VisJsRenderer(system, embeddable)
    return renderer.html_content()


def to_visjs(system: System, filename: str, embeddable=False) -> None:
    """Export a system as HTML using vis.JS library.

    Parameters
    ----------
    system : System
        System to export
    filename : str
        Filename to write to
    embeddable: bool, optional
        Is the HTML to be embedded in an existing page? Default: False
    """
    ext = os.path.splitext(filename)[1]
    html_ext = f"{os.extsep}html"
    if ext != html_ext:
        filename += html_ext

    rendered_html = visjs_html(system, embeddable)
    with open(filename, mode="w", encoding="utf-8") as f:
        f.write(rendered_html)



import logging

from typing import Optional
from cosapp.utils.validate import validate
from cosapp.drivers.optionaldriver import OptionalDriver

logger = logging.getLogger(__name__)


# TODO
# [ ] Quid for vector variables
class ValidityCheck(OptionalDriver):
    """
    When executed, this driver reports in the log the validity status for all variables of the
    driver `System` owner (and its children).

    Parameters
    ----------
    name : str
        Name of the driver
    owner : System
        :py:class:`~cosapp.systems.system.System` to which this driver belong
    **kwargs : Any
        Keyword arguments will be used to set driver options
    """
    
    __slots__ = tuple()

    def __init__(self, name: str, owner: Optional["cosapp.systems.System"] = None, **kwargs) -> None:
        """Initialize a driver

        Parameters
        ----------
        name: str
            Name of the `Driver`
        owner : System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belongs; default None
        **kwargs : Dict[str, Any]
            Optional keywords arguments
        """
        super().__init__(name, owner, **kwargs)

    def compute(self) -> None:
        """Report in the log the validity status for all variables
        recursively collected in owner `System` and its children.
        """
        warnings, errors = validate(self.owner)

        def message(log_dict):
            return "\n" + "\n\t".join(f"{key}{msg}" for key, msg in log_dict.items())

        if warnings:
            logger.warning(message(warnings))

        if errors:
            logger.error(message(errors))



import pytest

from cosapp.base import System
from cosapp.tests.library.systems import FanComplex, ComplexTurbofan
from cosapp.tools.views.visjs import VisJsRenderer
from cosapp.utils.testing import no_exception, assert_keys


pytest.importorskip("jinja2")


@pytest.fixture(scope="function")
def fan():
    return FanComplex("fan")


@pytest.fixture(scope="function")
def renderer(fan):
    return VisJsRenderer(fan)


def test_VisJsRenderer_html_tags(renderer):
    tags = renderer.html_tags()
    assert_keys(tags, "html_begin_tags", "html_end_tags")


def test_VisJsRenderer_html_resources(renderer):
    resources = renderer.html_resources()
    assert_keys(resources, "visJS", "visCSS")


def test_VisJsRenderer_html_template(renderer):
    with no_exception():
        template = renderer.html_template()


def test_VisJsRenderer_get_globals(renderer):

    common = renderer.get_globals()
    assert_keys(
        common, "template", "html_begin_tags", "html_end_tags", "visJS", "visCSS"
    )


def test_VisJsRenderer_html_content(renderer):
    with no_exception():
        renderer.html_content()


@pytest.mark.parametrize("name, expected", [
    ('foo', 'foo'),
    ('very_long_name', 'very_l…'),
])
def test_VisJsRenderer_get_system_label(name, expected):
    s = System(name)
    label = VisJsRenderer.get_system_label(s)
    assert label == expected


def test_VisJsRenderer_get_system_title():
    class Head(System):
        pass

    class Sub1(System):
        pass

    class Sub2(System):
        pass

    head = Head('head')
    sub1 = head.add_child(Sub1('sub1'))
    sub2 = sub1.add_child(Sub2('sub2'))
    visjs = VisJsRenderer(head)
    assert visjs.get_system_title(head) == "head - Head"
    assert visjs.get_system_title(sub1) == "sub1 - Sub1"
    assert visjs.get_system_title(sub2) == "sub1.sub2 - Sub2"

    # New renderer, based on a sub-system
    visjs = VisJsRenderer(sub1)
    assert visjs.get_system_title(sub1) == "sub1 - Sub1"
    assert visjs.get_system_title(sub2) == "sub2 - Sub2"
    with pytest.raises(ValueError, match="not a child"):
        visjs.get_system_title(head)


def test_VisJsRenderer_get_data():
    turbofan = ComplexTurbofan("turbofan")

    visjs = VisJsRenderer(turbofan)
    data = visjs.get_data()
    assert set(data) == {
        "title",
        "nodes",
        "edges",
        "groups",
    }
    assert set(data["groups"]) == {
        "turbofan",
        "turbofan.fanC",
        "turbofan.fanC.ductC",
    }
    assert set(node["title"] for node in data["nodes"]) == {
        # Systems:
        "turbofan - ComplexTurbofan",
        "atm - Atm",
        "inlet - Inlet",
        "fanC - FanComplex",
        "fanC.ductC - ComplexDuct",
        "fanC.ductC.merger - Merger",
        "fanC.ductC.duct - Duct",
        "fanC.ductC.bleed - Splitter",
        "fanC.fan - Fan",
        "merger - Merger",
        "duct - Duct",
        "bleed - Splitter",
        "noz - Nozzle",
        # Ports:
        "fanC.inwards",
        "fanC.fl_out - FluidPort",
        "fanC.fl_in - FluidPort",
        "fanC.mech_in - MechPort",
        "fanC.ductC.fl_out - FluidPort",
        "fanC.ductC.fl_in - FluidPort",
    }

    # Test on sub-system
    visjs = VisJsRenderer(turbofan.fanC)
    data = visjs.get_data()
    assert set(data) == {
        "title",
        "nodes",
        "edges",
        "groups",
    }
    assert set(data["groups"]) == {
        "fanC",
        "fanC.ductC",
    }
    assert set(node["title"] for node in data["nodes"]) == {
        # Systems:
        "fanC - FanComplex",
        "ductC - ComplexDuct",
        "ductC.merger - Merger",
        "ductC.duct - Duct",
        "ductC.bleed - Splitter",
        "fan - Fan",
        # Ports:
        "fanC.inwards",
        "fanC.fl_in - FluidPort",
        "fanC.fl_out - FluidPort",
        "fanC.mech_in - MechPort",
        "fanC.ductC.fl_out - FluidPort",
        "fanC.ductC.fl_in - FluidPort",
    }



import abc
import json
import numpy
import logging
import re
from copy import copy
from numbers import Number
from typing import (
    AnyStr, Dict, List,
    Callable, Optional,
    Sequence, Tuple, Union,
)

from cosapp.core.numerics.basics import MathematicalProblem, SolverResults
from cosapp.drivers.driver import Driver
from cosapp.drivers.runonce import RunOnce
from cosapp.utils.options_dictionary import OptionsDictionary

logger = logging.getLogger(__name__)


class AbstractSolver(Driver):
    """
    Solve a `System`

    Parameters
    ----------
    name : str
        Name of the driver
    owner : System, optional
        :py:class:`~cosapp.systems.system.System` to which driver belongs; defaults to `None`
    **kwargs : Any
        Keyword arguments will be used to set driver options
    """

    __slots__ = ('force_init', '_raw_problem', 'problem', 'initial_values', 'solution')

    def __init__(
        self,
        name: str,
        owner: Optional["cosapp.systems.System"] = None,
        **kwargs
    ) -> None:
        """Initialize driver

        Parameters
        ----------
        name: str, optional
            Name of the `Driver`.
        owner: System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; defaults to `None`.
        **kwargs:
            Additional keywords arguments forwarded to base class.
        """
        super().__init__(name, owner, **kwargs)

        self.force_init = False  # type: bool
            # desc="Force the initial values to be used."

        self.problem: MathematicalProblem = None
        self.initial_values = numpy.empty(0, dtype=float)
        self.solution: Dict[str, float] = {}
        self.reset_problem()

    def _set_owner(self, system: Optional["cosapp.systems.System"]) -> bool:
        defined = self.owner is not None
        changed = super()._set_owner(system)
        if changed:
            self.reset_problem()
            if defined:
                logger.warning(
                    f"System owner of Driver {self.name!r} has changed. Mathematical problem has been cleared."
                )
        return changed

    @property
    def raw_problem(self) -> MathematicalProblem:
        """MathematicalProblem: raw problem defined at solver level"""
        return self._raw_problem

    def reset_problem(self) -> None:
        """Reset mathematical problem"""
        self._raw_problem = MathematicalProblem(self.name, self.owner)

    def _filter_options(self, kwargs, aliases: Dict[str, str] = dict()):
        """
        Translate option names into self.options using an alias dictionary, to handle cases where
        a common option name, such as 'tol', is passed to a specific solver/function with a different name.
        
        For example, in scipy.optimize.root(), the convergence criterion may be referred to as 'ftol', 'gtol'...
        depending on the invoked algorithm (Levenberg-Marquardt, Powell, Broyden's good, etc.).
        """
        keys = list(kwargs.keys())  # make a copy of keys only

        for key in keys:
            value = kwargs.pop(key)
            try:
                key = aliases[key]
            except KeyError:
                pass
            if key in self.options:
                self.options[key] = value
            else:
                raise KeyError(f"Unknown solver option {key!r}")

    def _get_solver_limits(self) -> Dict[str, numpy.ndarray]:
        """Returns the step limitations for all iteratives.

        There are 4 types of limits defined:
        - lower_bound: lower bound of the iteratives
        - upper_bound: upper bound of the iteratives
        - abs_step: maximal absolute step of the iteratives
        - rel_step: maximal relative step of the iteratives

        Returns
        -------
        Dict[str, numpy.ndarray]
            Dictionary with the limits for all iteratives.
        """
        mapping = {
            "lower_bound": "lower_bound",
            "upper_bound": "upper_bound",
            "abs_step": "max_abs_step",
            "rel_step": "max_rel_step",
        }
        options = dict.fromkeys(mapping, [])

        for unknown in self.problem.unknowns.values():
            for name, attr_name in mapping.items():
                const = getattr(unknown, attr_name)
                array = numpy.full_like(unknown.value, const, dtype=type(const))
                options[name] = numpy.concatenate((options[name], array.flatten()))

        return options

    def compute_before(self):
        """Contains the customized `Module` calculation, to execute before children.
        """
        # Set initial_values to current solution
        # k = 0
        # for value in self.solution.values():
        #     n = numpy.size(value)
        #     self.initial_values[k : k + n] = numpy.reshape(value, -1)
        #     k += n
        logger.debug(f"Set unknowns initial values: {self.initial_values}")
        self.set_iteratives(self.initial_values)

    def setup_run(self) -> None:
        """Set up the mathematical problem."""
        super().setup_run()
        self.problem = MathematicalProblem(self.name, self.owner)
        self.initial_values = numpy.empty(0, dtype=float)

    def run_once(self) -> None:
        """Run solver once, assuming driver has already been initialized.
        """
        with self.log_context(" - run_once"):
            if self.is_active():
                self._precompute()

                logger.debug(f"Call {self.name}.compute_before()")
                self.compute_before()

                # Sub-drivers are executed at each iteration in `compute`,
                # so the child loop before `self.compute()` is omitted.
                logger.debug(f"Call {self.name}.compute()")
                self._compute_calls += 1
                self.compute()

                self._postcompute()
                self.computed.emit()
            
            else:
                logger.debug(f"Skip {self.name} execution - Inactive")

    def _precompute(self) -> None:
        # TODO we should check that all variables are of numerical types
        super()._precompute()
        self.touch_unknowns()

    def touch_unknowns(self):
        for unknown in self.problem.unknowns.values():
            unknown.touch()

    @abc.abstractmethod
    def set_iteratives(self, x: Sequence[float]) -> None:
        pass

    @abc.abstractmethod
    def resolution_method(self,
        fresidues: Callable[[Sequence[float], Union[float, str], bool], numpy.ndarray],
        x0: Sequence[float],
        args: Tuple[Union[float, str]] = (),
        options: Optional[OptionsDictionary] = None,
    ) -> SolverResults:
        """Function call to cancel the residues.

        Parameters
        ----------
        fresidues : Callable[[Sequence[float], Union[float, str]], numpy.ndarray]
            Residues function taking two parameters (evaluation vector, time/ref) and returning the residues
        x0 : Sequence[float]
            The initial values vector to converge to the solution
        args : Tuple[Union[float, str], bool], optional
            A tuple of additional argument for fresidues starting with the time/ref parameter and the need to update
            residues reference
        options : OptionsDictionary, optional
            Options for the numerical resolution method

        Returns
        -------
        SolverResults
            Solution container
        """
        pass

    # def _postcompute(self) -> None:
    #     # Set initial_values to current solution
    #     k = 0
    #     for value in self.solution.values():
    #         n = numpy.size(value)
    #         self.initial_values[k : k + n] = numpy.reshape(value, -1)
    #         k += n
    #     super()._postcompute()

    def save_solution(self, file: Optional[str] = None) -> Dict[str, Union[Number, List[Number]]]:
        """Save the latest solver solution.

        If `file` is specified, the solution will be saved in it in JSON format.

        Parameters
        ----------
        file : str, optional
            Filename to save the answer in; default None (i.e. data will not be saved)

        Returns
        -------
        Dict[str, Union[Number, List[Number]]]
            Dictionary of the latest solution
        """
        latest_answer = dict()

        for k, v in self.solution.items():
            if isinstance(v, numpy.ndarray):
                v = v.tolist()
            latest_answer[k] = copy(v)

        if file:
            with open(file, "w") as outfile:
                json.dump(latest_answer, outfile)

        return latest_answer

    def load_solution(self,
        solution: Union[Dict[str, Union[Number, numpy.ndarray]], AnyStr],
        case: Optional[str] = None,
    ):
        """Load the provided solution to initialize the solver.

        The solution can be provided directly as a dictionary or from a filename to be read.

        Parameters
        ----------
        solution : Dict[str, Union[Number, numpy.ndarray]] or str
            Dictionary of the latest solution to load or the filename in JSON format to read from.
        case : str, optional
            Case to initialize with the solution; default None (i.e. will be guessed from variable name)
        """
        # TODO Fred is it better to set the initial values or to override the previous solution?
        #   In case the solution does not cover all offdesign unknowns, the later should be better.
        from cosapp.systems import System

        if isinstance(solution, str):
            with open(solution, "r") as f:
                data = json.load(f)
        elif isinstance(solution, dict):
            data = solution
        else:
            raise TypeError(
                f"Solution expected as dict or json file name; got {type(solution).__qualname__!r}."
            )

        def extract_varname(driver: Driver, key: str):
            matches = re.findall(f"{driver.name}\[(.*)\]", key)
            if matches:  # Off-design variable
                return matches[0]
            else:
                return key

        def is_RunOnce(driver: Driver) -> bool:
            return isinstance(driver, RunOnce)

        with System.set_master(repr(self.owner)) as is_master:
            if is_master:
                self.owner.open_loops()

            try:
                if case is None:
                    for name, value in data.items():
                        for child in filter(is_RunOnce, self.children.values()):
                            varname = extract_varname(child, name)
                            if varname != name:  # Off-design variable
                                try:
                                    child.set_init({varname: numpy.asarray(value)})
                                except:
                                    continue
                                else:
                                    break
                            elif varname in child.design.unknowns:  # We may have a design variable
                                child.set_init({varname: numpy.asarray(value)})
                                break
                else:
                    child = self.children[case]
                    if not is_RunOnce(child):
                        raise TypeError(
                            "Only drivers derived from RunOnce can be initialized"
                            f"; got {type(child).__qualname__!r} for driver {case!r}."
                        )
                    for name, value in data.items():
                        varname = extract_varname(child, name)
                        try:
                            child.set_init({varname: numpy.asarray(value)})
                        except:
                            continue

            finally:  # Ensure to clean the system
                if is_master:
                    self.owner.close_loops()



import logging
import copy
from collections import OrderedDict
from typing import Any, Dict, Iterator, Generator, Optional, Tuple, Union, Callable, TypeVar
from types import MappingProxyType

from cosapp.patterns import visitor
from cosapp.ports.enum import PortType, Scope, Validity
from cosapp.ports.exceptions import ScopeError
from cosapp.ports.variable import RangeValue, Types, BaseVariable, Variable
from cosapp.ports.mode_variable import ModeVariable
from cosapp.utils.distributions import Distribution
from cosapp.utils.helpers import check_arg
from cosapp.utils.naming import NameChecker

logger = logging.getLogger(__name__)

T = TypeVar("T")


class BasePort(visitor.Component):
    """Base class for ports, containers gathering variables.

    Common users should not use this class directly.

    Parameters
    ----------
    name : str
        Port name
    direction : {PortType.IN, PortType.OUT}
        Port direction
    """

    _name_check = NameChecker()

    def __init__(self, name: str, direction: PortType) -> None:
        """`BasePort` constructor.

        Parameters
        ----------
        name : str
            Port name
        direction : {PortType.IN, PortType.OUT}
            Port direction
        """
        if not isinstance(direction, PortType):
            raise TypeError(f"Direction must be PortType; got {direction}.")
        from cosapp.systems import System

        self.__is_clean = False
        self._variables: Dict[str, BaseVariable] = OrderedDict()
        self._name: str = self._name_check(name)
        self._desc: str = ""
        self._direction: PortType = direction
        self._owner: Optional[System] = None
        self.__clearance: Scope = None
        self.scope_clearance = Scope.PRIVATE

    def accept(self, visitor: visitor.Visitor) -> None:
        """Specifies course of action when visited by `visitor`"""
        visitor.visit_port(self)

    @property
    def owner(self) -> Optional["cosapp.systems.System"]:
        """System : `System` owning the port."""
        return self._owner

    @owner.setter
    def owner(self, new_owner: "cosapp.systems.System"):
        from cosapp.systems import System

        if not isinstance(new_owner, System):
            raise TypeError(f"Port owner must be a `System`; got {new_owner!r}.")
        self._owner = new_owner
        self.__update_filter()

    @property
    def name(self) -> str:
        """str : Port name"""
        return self._name

    @property
    def contextual_name(self) -> str:
        """str : Join port owner name and port name.

        If the port has no owner, the port name is returned.
        """
        owner = self._owner
        return self._name if owner is None else f"{owner.name}.{self._name}"

    def full_name(self, trim_root=False) -> str:
        """Returns full name up to root owner.
        
        Parameters
        ----------
        trim_root : bool (optional, default False)
            Exclude root owner name if True.

        Returns
        -------
        str
            The port full name
        """
        owner = self.owner
        path = []
        if owner is not None:
            path = owner.path_namelist()
            if trim_root:
                path = path[1:]
        path.append(self.name)
        return ".".join(path)

    @property
    def description(self) -> str:
        """str: Port description"""
        return self._desc

    @description.setter
    def description(self, desc: str) -> None:
        check_arg(desc, 'description', str)
        self._desc = desc

    @property
    def direction(self) -> PortType:
        """:obj:`PortType.IN` or :obj:`PortType.OUT` : Port direction"""
        return self._direction

    @property
    def is_input(self) -> bool:
        """bool: True if port is an input, False otherwise."""
        return self._direction == PortType.IN

    @property
    def is_output(self) -> bool:
        """bool: True if port is an output, False otherwise."""
        return self._direction == PortType.OUT

    @property
    def is_clean(self) -> bool:
        return self.__is_clean

    def set_clean(self) -> None:
        """Set port as 'clean'."""
        self.__is_clean = True

    def touch(self) -> None:
        """Set port as 'dirty'."""
        self.__is_clean = False
        if self._owner and self.is_input:
            self._owner.touch()

    def __repr__(self) -> str:
        return f"{type(self).__name__}: {self.serialize_data()!r}"

    def _repr_markdown_(self) -> str:
        """Returns the representation of this port variables in Markdown format.

        Returns
        -------
        str
            Markdown formatted representation
        """
        from cosapp.tools.views.markdown import port_to_md

        return port_to_md(self)

    def __json__(self) -> Dict[str, Dict[str, Any]]:
        """JSONable dictionary representing a variable.
        
        Returns
        -------
        Dict[str, Any]
            The dictionary
        """
        return dict(
            (name, variable.__json__())
            for name, variable in self._variables.items()
        )

    def serialize_data(self) -> Dict[str, Any]:
        """Serialize the variable values in a dictionary.
        
        Returns
        -------
        Dict[str, Any]
            The dictionary (variable name, value)
        """
        return dict(self.items())

    def add_variable(
        self,
        name: str,
        value: Any = 1,
        unit: str = "",
        dtype: Types = None,
        valid_range: RangeValue = None,
        invalid_comment: str = "",
        limits: RangeValue = None,
        out_of_limits_comment: str = "",
        desc: str = "",
        distribution: Optional[Distribution] = None,
        scope: Scope = Scope.PRIVATE,
    ) -> None:
        """Add a variable to the port.

        The `valid_range` defines the range of value for which a model is known to behave correctly.
        The `limits` are at least as large as the validity range.

        Parameters
        ----------
        name : str
            Name of the variable
        value : Any, optional
            Value of the variable; default 1
        unit : str, optional
            Variable unit; default empty string (i.e. dimensionless)
        dtype : type or iterable of type, optional
            Variable type; default None (i.e. type of initial value)
        valid_range : Tuple[Any, Any] or Tuple[Tuple], optional
            Validity range of the variable; default None (i.e. all values are valid)
        invalid_comment : str, optional
            Comment to show in case the value is not valid; default ''
        limits : Tuple[Any, Any] or Tuple[Tuple], optional
            Limits over which the use of the model is wrong; default valid_range
        out_of_limits_comment : str, optional
            Comment to show in case the value is not valid; default ''
        desc : str, optional
            Variable description; default ''
        distribution : Distribution, optional
            Probability distribution of the variable; default None (no distribution)
        scope : Scope {PRIVATE, PROTECTED, PUBLIC}, optional
            Variable visibility; default PRIVATE
        """
        if name in self:
            logger.warning(
                f"Variable {name} already exists in port {self.contextual_name}."
                " It will be overwritten."
            )

        # Value must be set before creating variable as validation in Variable needs it
        self._variables[name] = variable = Variable(
            name,
            self,
            value,
            unit,
            dtype,
            valid_range,
            invalid_comment,
            limits,
            out_of_limits_comment,
            desc,
            distribution,
            scope,
        )
        value = variable.filter_value(value)

        # For efficiency reasons, values are stored as port attributes
        setattr(self, name, value)

    # TODO should we override __setattr__ to forward setting to the source port? and/or raise an
    #   error if it is not possible - for example if the source is an output variable
    def __set_notype_checking(self, key, value):
        super().__setattr__(key, value)

        if self.__is_clean and key in self._variables:
            self.touch()

    def validate(self, key: str, value: Any) -> None:
        """Check if a variable is in the scope of the user and the type is valid.
        
        Parameters
        ----------
        key : str
            Name of the variable to test
        value : Any
            Value to validate

        Raises
        ------
        ScopeError
            If the variable is not visible for the user
        TypeError
            If the value has an unauthorized type
        """
        # Check scope
        if self.__out_of_scope(key):
            raise ScopeError(f"Cannot set out-of-scope variable {key!r}.")

        if value is not None:
            var_type = self._variables[key].dtype
            ok = var_type is None or isinstance(value, var_type)
            if not ok:
                raise TypeError(
                    "Trying to set {}.{} of type {} with {}.".format(
                        self.contextual_name, key, var_type, type(value)
                    )
                )

    def __set_variable(self, key, value):
        """Set the variable `key` with `value`.
        
        Parameters
        ----------
        key : str
            Name of the variable to be set
        value : Any
            Value to set
        """
        if key.startswith("_"):
            super().__setattr__(key, value)
        elif key in self._variables:
            self.validate(key, value)
            self.__set_notype_checking(key, value)
        elif hasattr(self, key):
            super().__setattr__(key, value)
        else:
            raise AttributeError(
                f"Port variable {self.contextual_name}.{key} can only be created using method 'add_variable'."
            )

    __setattr__ = __set_variable

    @staticmethod
    def set_type_checking(activate: bool) -> None:
        """(Un)set type checking when affecting port variables.

        By default type checking is activated.

        Parameters
        ----------
        activate : bool
            True to activate type checking, False to deactivate
        """
        if activate:
            BasePort.__setattr__ = BasePort.__set_variable
        else:
            BasePort.__setattr__ = BasePort.__set_notype_checking

    def __contains__(self, item: str) -> bool:
        return item in self._variables

    def __getitem__(self, item: str) -> Any:
        try:
            return getattr(self, item)
        except AttributeError:
            raise KeyError(f"Variable or property {item} does not exist in Port {self}.")

    def __setitem__(self, key: str, value: Any) -> None:
        if key in self._variables:
            setattr(self, key, value)
        else:
            raise KeyError(f"Variable {key} does not exist in Port {self}.")

    def __iter__(self) -> Iterator[str]:
        return iter(self._variables)

    def __len__(self) -> int:
        return len(self._variables)

    def items(self) -> Generator[Tuple[str, Any], None, None]:
        """Dictionary-like item generator yielding (name, value) tuples
        for all port variables."""
        for name in self._variables:
            yield (name, getattr(self, name))

    def set_values(self, **modifications) -> None:
        """Generic setter for port variable values, offering a
        convenient way of modifying multiple variables at once.

        Parameters:
        -----------
        **modifications:
            Variable names and values given as keyword arguments.
        
        Examples:
        ---------
        >>> from cosapp.base import Port, System
        >>>
        >>> class DummyPort(Port):
        >>>     def setup(self):
        >>>         self.add_variable('a')
        >>>         self.add_variable('b')
        >>>
        >>> class DummySystem(System):
        >>>     def setup(self):
        >>>         self.add_input(DummyPort, 'p_in')
        >>>         self.add_output(DummyPort, 'p_out')
        >>>
        >>>     def compute(self):
        >>>         p_in = self.p_in
        >>>         self.p_out.set_values(
        >>>             a = p_in.b,
        >>>             b = p_in.a - p_in.b,
        >>>         )
        """
        for name, value in modifications.items():
            setattr(self, name, value)

    @property
    def scope_clearance(self) -> Scope:
        """Scope: Current clearance level of the port.
        Determines the set of read-only port variables."""
        return self.__clearance

    @scope_clearance.setter
    def scope_clearance(self, user_scope: Scope) -> None:
        check_arg(user_scope, "scope_clearance", Scope)
        self.__clearance = user_scope
        self.__update_filter()

    def __update_filter(self) -> None:
        """Update the implementation of method `out_of_scope`
        according to port's clearance level and owner."""
        if self._owner is None:
            criterion = lambda name: False
        else:
            read_only = [
                name for (name, variable) in self._variables.items()
                if variable.scope > self.__clearance
            ]
            criterion = lambda name: not self._owner.is_running() and name in read_only
        self.__out_of_scope = criterion

    def out_of_scope(self, name: str) -> bool:
        """
        Asserts if current scope `scope_clearance` is high enough to
        allow the modification of port variable `name`.

        Parameters
        ----------
        name : str
            Variable name to test

        Returns
        -------
        bool
            Is modification forbidden?

        The behavior of this function is set by method `scope_clearance()`.
        By default, clearance level is set to `Scope.PUBLIC`.

        Examples
        --------
        >>> port = Port("myPort", PortType.IN)
        >>> port.add_variable("x", 1.5, scope=Scope.PUBLIC)
        >>> port.add_variable("y", 0.2, scope=Scope.PROTECTED)
        >>> port.add_variable("z", 0.3, scope=Scope.PRIVATE)
        >>>
        >>> port.scope_clearance(Scope.PROTECTED)  # only `port.x` and `port.y` can be modified
        >>> assert not port.out_of_scope("x")
        >>> assert not port.out_of_scope("y")
        >>> assert port.out_of_scope("z")
        >>>
        >>> port.scope_clearance(Scope.PUBLIC)     # only `port.x` can be modified
        >>> assert not port.out_of_scope("x")
        >>> assert port.out_of_scope("y")
        >>> assert port.out_of_scope("z")
        """
        return self.__out_of_scope(name)

    def get_details(self, name: Optional[str] = None) -> Union[Dict[str, BaseVariable], BaseVariable]:
        """Return the variable(s).
        
        Parameters
        ----------
        name : str, optional
            Name of the variable looked for; default None (all variable are returned).

        Returns
        -------
        types.MappingProxyType[str, cosapp.ports.variable.Variable] or cosapp.ports.variable.Variable
            The sought variable, or a read-only view on all port variables.
        """
        if name is None:
            return MappingProxyType(self._variables)
        else:
            return self._variables[name]

    def variables(self) -> Iterator[BaseVariable]:
        """Iterator over port `BaseVariable` instances."""
        return self._variables.values()

    def check(self, name: Optional[str] = None) -> Union[Dict[str, Validity], Validity]:
        """Get the variable value validity.

        If `name` is not provided, returns a dictionary with the validity of all variables. Else
        only, the validity for the given variable will be returned.

        Parameters
        ----------
        name : str, optional
            Variable name; default None = All variables will be tested

        Returns
        -------
        Dict[str, Validity] or Validity
            (Dictionary of) the variable(s) value validity
        """
      
        if name is None:
            return dict(
                (name, variable.is_valid())
                for name, variable in self._variables.items()
            )
        else:           
            return self._variables[name].is_valid()

    def get_validity_ground(self,
        status: Validity,
        name: Optional[str] = None,
    ) -> Union[Dict[str, str], str]:
        # TODO unit tests
        """Get the ground arguments used to established the variable validity range.

        The status `Validity.OK` has no ground arguments.

        Parameters
        ----------
        status : Validity.{OK, WARNING, ERROR}
            Validity status for which the reasons are looked for.
        name : str, optional
            Variable name; default None = All variables will be tested

        Returns
        -------
        Dict[str, str] or str
            (Dictionary of) the variable(s) validity ground
        """
        if name is None:
            return dict(
                (name, variable.get_validity_comment(status))
                for name, variable in self._variables.items()
            )
        else:
            return self._variables[name].get_validity_comment(status)

    def copy(
        self,
        name: Optional[str] = None,
        direction: Optional[PortType] = None,
    ):
        """Duplicates the port.

        Parameters
        ----------
        name : str, optional
            Name of the duplicated port; default original port name
        direction : {PortType.IN, PortType.OUT}, optional
            Direction of the duplicated port; default original direction

        Returns
        -------
        Port
            The copy of the current port
        """
        cls = type(self)
        port = cls(
            name or self.name,
            direction or self.direction,
        )
        for name in self._variables:
            port.copy_variable_from(self, name)

        return port

    def copy_variable_from(self, port: "BasePort", name: str, alias: Optional[str] = None) -> None:
        """Copy variable `name` from another port into variable `alias`.

        Parameters
        ----------
        port : BasePort
            Port from which variable is copied
        name : str
            Variable name from source port.
        alias : str, optional
            Variable name in current port (same as `name` if not specified).
        """
        if alias is None:
            alias = name
        variable = port._variables[name]
        value = getattr(port, name)
        self._variables[alias] = variable.copy(self, alias)
        setattr(self, alias, copy.copy(value))

    def morph(self, port: "BasePort") -> None:
        """Morph the provided port into this port.

        Morphing a port is useful when converting a `System` in something equivalent. The morphing
        goal is to preserve connections and adapt the port content if needed.

        Parameters
        ----------
        port : BasePort
            The port to morph to
        """

        # TODO unit test for variable details when morphing
        for name, variable in self._variables.items():
            if name not in port:
                port.add_variable(
                    name,
                    getattr(self, name),
                    unit=variable.unit,
                    dtype=variable.dtype,
                    desc=variable.description,
                    valid_range=variable.valid_range,
                    invalid_comment=variable.invalid_comment,
                    limits=variable.limits,
                    out_of_limits_comment=variable.out_of_limits_comment,
                    distribution=variable.distribution,
                    scope=variable.scope,
                )

        for variable in list(port):
            if variable not in self:
                port.remove_variable(variable)

    def to_dict(self, with_def: bool = False) -> Dict[str, Union[str, Tuple[Dict[str, str], str]]]:
        """Convert this port in a dictionary.
   
        Parameters
        ----------
        with_def : bool
            Flag to export also output ports and its class name (default: False).

        Returns
        -------
        dict
            The dictionary representing this port.
        """
        # TODO this is uncomplete as validation ranges and distribution could be changed
        new_dict = dict()

        if with_def:
            data = dict()
            if self.name not in ["inwards", "outwards"]:
                data["__class__"] = self.__class__.__qualname__
                data.update(self.items())
            else:
                data.update(
                    (varname, variable.to_dict())
                    for varname, variable in self._variables.items()
                )
            new_dict[self.name] = data
        
        elif self.is_input:
            portname = self.name
            new_dict.update(
                (f"{portname}.{varname}", value)
                for varname, value in self.items()
            )
        
        return new_dict


class ModeVarPort(BasePort):
    """Class used for the local storage of mode variables."""

    # TODO: Forbid the use of add_variable
    # TODO: Redefine exports/serializations
    # TODO: Typing of _variables is ambiguous

    def add_mode_variable(
        self,
        name: str,
        value: Optional[Any] = None,
        unit: str = "",
        dtype: Types = None,
        desc: str = "",
        init: Optional[Any] = None,
        scope: Scope = Scope.PRIVATE,
    ) -> None:
        """Add a mode variable to the port.

        Parameters
        ----------
        name : str
            Name of the variable
        value : Any, optional
            Value of the variable; default 1
        unit : str, optional
            Variable unit; default empty string (i.e. dimensionless)
        dtype : type or iterable of type, optional
            Variable type; default None (i.e. type of initial value)
        desc : str, optional
            Variable description; default ''
        scope : Scope {PRIVATE, PROTECTED, PUBLIC}, optional
            Variable visibility; default PRIVATE
        """
        if name in self:
            logger.warning(
                f"Variable {name} already exists in multimode port {self.contextual_name}."
                " It will be overwritten."
            )
        # Value must be set before creating variable as validation in Variable needs it
        if init is not None and value is not None and self.is_input:
            logger.warning(
                f"Initial value {init} is discarded for input mode variable {name!r}"
            )
            init = None
        
        self._variables[name] = variable = ModeVariable(
            name = name,
            port = self,
            value = value,
            unit = unit,
            dtype = dtype,
            desc = desc,
            init = init,
            scope = scope,
        )
        if value is None:
            value = variable.init_value()
        value = variable.filter_value(value)

        # For efficiency reasons, variables are stored as port attributes
        setattr(self, name, value)


class ExtensiblePort(BasePort):
    """Class describing ports with a varying number of variables."""

    # TODO unused and should be removed -> to dangerous for consistency
    def remove_variable(self, name: str) -> None:
        """Removes a variable from the port.

        Parameters
        ----------
        name : str
            Name of the variable to be removed

        Raises
        ------
        AttributeError
            If the variable does not exists
        """
        if name not in self._variables:
            raise AttributeError(
                f"Variable {name!r} does not exist in port {self.contextual_name}"
            )
        delattr(self, name)
        self._variables.pop(name)


class Port(BasePort):
    """A `Port` is a container gathering variables tightly linked.

    An optional dictionary may be specified to overwrite variable value and some of their
    metadata. Values of the dictionary may be of two kinds:
    - Only the new value
    - A dictionary with new value and/or new details

    If the value or a metadata is set to `None`, the default value will be kept.

    Parameters
    ----------
    name : str
        `Port` name
    direction : {PortType.IN, PortType.OUT}
        `Port` direction
    variables : Dict[str, Any], optional
        Dictionary of variables with their value and details; default: None = default value

    Attributes
    ----------
    _locked : bool
        if True, `add_variable` is deactivated. This is the default behavior outside the `setup`
        function.

    Examples
    --------

    Use this class by subclassing it:

    >>> class FlowPort(Port):
    >>>     def setup(self):
    >>>         self.add_variable('Pt', 101325., unit='Pa')
    >>>         self.add_variable('Tt', 273.15, unit='K')
    >>>         self.add_variable('W', 1.0, unit='kg/s')
    >>>
    >>> p = FlowPort('myPort', PortType.IN)
    >>> # Overwrite value and some details
    >>> f = FlowPort(
    >>>     'myPort2', 
    >>>     PortType.OUT,
    >>>     {
    >>>         'Pt': 10e6,
    >>>         'W': {
    >>>             'value': 10.,  # New value
    >>>             'unit': 'lbm/s',  # New unit - should be compatible of the original one
    >>>             'valid_range': (0., None),  # Updated validated range
    >>>             'invalid_comment': 'Flow should be positive',  # Comment if out of validated range
    >>>             'limits': None,  # Limits
    >>>             'out_of_limits_comment': '',  # Comment if out of limits
    >>>         }
    >>>     }
    >>> )
    """

    def __init__(
        self,
        name: str,
        direction: PortType,
        variables: Optional[Dict[str, Any]] = None,
    ) -> None:
        """`Port` constructor.

        An optional dictionary may be specified to overwrite variable value and some of their
        metadata. Values of the dictionary may be of two kinds:
        - Only the new value
        - A dictionary with new value and/or new details

        Parameters
        ----------
        name : str
            `Port` name
        direction : {PortType.IN, PortType.OUT}
            `Port` direction
        variables : Dict[str, Any], optional
            Dictionary of variables with their value and details; default: None = default value
        """
        super().__init__(name, direction)
        self.__locked = False  # type: bool

        self.setup()

        if variables is not None:
            check_arg(variables, "variables", dict)
            for name, value in variables.items():
                variable_value = value
                if isinstance(value, dict):
                    variable_value = value.pop("value", None)

                    details = self.get_details(name)
                    try:
                        for field, param in value.items():
                            setattr(details, field, param)
                    except AttributeError:  # Unexpected keyword for details
                        # If current variable is of type dict
                        if isinstance(self[name], dict):
                            if variable_value is not None:
                                value["value"] = variable_value
                            variable_value = value
                        else:
                            raise

                if variable_value is not None:
                    self[name] = variable_value

        self.__locked = True

    def setup(self) -> None:
        """`Port` variables are defined in this function by calling `add_variable`.

        This function allows to populate a customized `Port` class. The `add_variable`
        function is only callable in the frame of this function.

        Examples
        --------

        Here is an example of `Port` subclassing:

        >>> class FlowPort(Port):
        >>>     def setup(self):
        >>>         self.add_variable('Pt', 101325.)
        >>>         self.add_variable('Tt', 273.15)
        >>>         self.add_variable('W', 1.0)
        """
        pass  # pragma: no cover

    def add_variable(
        self,
        name: str,
        value: Any = 1,
        unit: str = "",
        dtype: Types = None,
        valid_range: RangeValue = None,
        invalid_comment: str = "",
        limits: RangeValue = None,
        out_of_limits_comment: str = "",
        desc: str = "",
        distribution: Optional[Distribution] = None,
        scope=Scope.PUBLIC,
    ) -> None:
        """Add a variable to the port.

        The `valid_range` defines the range of value for which a model is known to behave correctly.
        The `limits` are at least as large as the validity range.

        Notes
        -----
        The default visibility of a `Port` variable is `Scope.PUBLIC`.
        This is a difference compared to `BasePort`.

        Parameters
        ----------
        name : str
            Name of the variable
        value : Any, optional
            Value of the variable; default 1
        unit : str, optional
            Variable unit; default empty string (i.e. dimensionless)
        dtype : type or iterable of type, optional
            Variable type; default None (i.e. type of initial value)
        valid_range : Tuple[Any, Any] or Tuple[Tuple], optional
            Validity range of the variable; default None (i.e. all values are valid)
        invalid_comment : str, optional
            Comment to show in case the value is not valid; default ''
        limits : Tuple[Any, Any] or Tuple[Tuple], optional
            Limits over which the use of the model is wrong; default valid_range
        out_of_limits_comment : str, optional
            Comment to show in case the value is not valid; default ''
        desc : str, optional
            Variable description; default ''        
        distribution : Distribution, optional
            Probability distribution of the variable; default None (no distribution)
        scope : Scope {PRIVATE, PROTECTED, PUBLIC}, optional
            Variable visibility; default PUBLIC
        """
        if self.__locked:
            raise AttributeError("add_variable cannot be called outside `setup`.")

        super().add_variable(
            name,
            value=value,
            unit=unit,
            dtype=dtype,
            valid_range=valid_range,
            invalid_comment=invalid_comment,
            limits=limits,
            out_of_limits_comment=out_of_limits_comment,
            desc=desc,
            distribution=distribution,
            scope=scope,
        )

    def copy(
        self, 
        name: Optional[str] = None,
        direction: Optional[PortType] = None,
    ):
        """Duplicates the port.

        Parameters
        ----------
        name : str, optional
            Name of the duplicated port; default original port name
        direction : {PortType.IN, PortType.OUT}, optional
            Direction of the duplicated port; default original direction

        Returns
        -------
        Port
            The copy of the current port
        """
        cls = type(self)
        new_port = cls(
            name or self.name,
            direction or self.direction,
        )
        new_port.set_from(self, copy.copy)
        return new_port

    def set_from(self,
        source: BasePort,
        transfer: Callable[[T], T] = lambda x: x,
        check_names: bool = True,
    ) -> None:
        """Set values from another port.

        Parameters:
        -----------
        - source [BasePort]:
            Source port.
        - transfer [Callable[[T], T], optional]:
            Transfer function to pass values. By default, `transfer` is the identity
            function, which corresponds to plain assignment target.var = source.var.
            Copies can be performed by setting `transfer = copy.copy`, e.g.
        - check_names [bool, optional]:
            If `True` (default), figure out common variables before transfering values.
            If current and source ports are of the same type, no check is performed.
        
        Examples:
        ---------
        >>> from cosapp.base import Port, System
        >>> import copy
        >>>
        >>> class DummyPort(Port):
        >>>     def setup(self):
        >>>         self.add_variable('a')
        >>>         self.add_variable('b')
        >>>
        >>> class DummySystem(System):
        >>>     def setup(self):
        >>>         self.add_inward('a')
        >>>         self.add_inward('x')
        >>>         self.add_input(DummyPort, 'p_in')
        >>>         self.add_output(DummyPort, 'p_out')
        >>>
        >>>     def compute(self):
        >>>         self.p_out.set_from(self.p_in)     # peer-to-peer: no check
        >>>         self.p_out.set_from(self.inwards)  # will transfer `self.a`
        >>>         # Peer-to-peer, with deepcopy:
        >>>         self.p_out.set_from(self.p_in, copy.deepcopy)
        >>>         # Transfer `inwards` into `p_out`, with no name check:
        >>>         # raises AttributeError, as `DummyPort` has no variable `x`
        >>>         self.p_out.set_from(
        >>>             self.inwards,
        >>>             check_names=False,
        >>>         )
        """
        peers = type(source) is type(self)  # most likely usage

        if peers or not check_names:  # faster
            common = iter(source) 
        else:
            common = set(source).intersection(self)
            if not common:
                logger.warning(
                    f"{self.contextual_name!r} and {source.contextual_name!r} have no common variables."
                )
        for name in common:
            value = getattr(source, name)
            setattr(self, name, transfer(value))



import pytest
import numpy as np

from .conftest import (
    PointMass, PointMassSolution,
    DampedMassSpring, HarmonicOscillatorSolution,
    ScalarOde,
)
from cosapp.drivers import RungeKutta
from cosapp.drivers.time.utils import SystemInterpolator


def test_SystemInterpolator_massSpring():
    t0, t1 = 0, 10
    spring = DampedMassSpring('spring')
    driver = spring.add_driver(RungeKutta(time_interval=(t0, t1)))

    x0, v0 = 0.25, 0.3
    exact = HarmonicOscillatorSolution(spring, x0, v0)

    view = SystemInterpolator(driver)
    view.interp = {
        'x': exact.get_function('x'),
        'v': exact.get_function('v'),
    }
    time = np.linspace(t0, t1, 11)
    for t in time:
        view.exec(t)
        assert spring.x == exact.x(t)
        assert spring.v == exact.v(t)
        assert spring.a == exact.a(t)


def test_SystemInterpolator_pointMass():
    t0, t1 = 0, 10
    p = PointMass('p')
    driver = p.add_driver(RungeKutta(time_interval=(t0, t1)))

    x0 = [-1., 0., 10]
    v0 = [8, 0, 9.5]
    exact = PointMassSolution(p, v0, x0)

    view = SystemInterpolator(driver)
    view.interp = {
        'x': exact.get_function('x'),
        'v': exact.get_function('v'),
    }
    time = np.linspace(t0, t1, 11)
    for t in time:
        view.exec(t)
        assert p.x == pytest.approx(exact.x(t), abs=0)
        assert p.v == pytest.approx(exact.v(t), abs=0)
        assert p.a == pytest.approx(exact.a(t), abs=0)


def test_SystemInterpolator_ode_scenario():
    """Test """
    t0, t1 = 0, 1
    ode = ScalarOde('ode')
    driver = ode.add_driver(RungeKutta(time_interval=(t0, t1)))
    driver.set_scenario(
        values = {
            'df': '0.1 * exp(-t / 5)',
        }
    )
    view = SystemInterpolator(driver)
    view.interp = {
        'f': lambda t: t,  # not the actual solution
    }
    time = np.linspace(t0, t1, 11)
    for t in time:
        view.exec(t)
        assert ode.df == 0.1 * np.exp(-t / 5)
        assert ode.f == t



import logging
from typing import Any, Union, Collection, Dict

logger = logging.getLogger(__name__)


def pull_variables(
    child: "cosapp.systems.System", 
    pulling: Union[str, Collection[str], Dict[str, str]],
) -> None:
    """Pull variables from child to the parent.

    Parameters
    ----------
    child: System
        `System` asking to pull variables to its parent
    pulling: str or List[str] or Dict[str, str]
        Map of child ports to pulled ports at the parent system level
    """
    from cosapp.systems import System
    from cosapp.ports.port import BasePort, Port
    from cosapp.utils.naming import CommonPorts

    parent: System = child.parent
    if parent is None:
        raise AttributeError(
            f"Can't pull variables from orphan System {child.name!r}"
        )
    method_map = {
        CommonPorts.INWARDS.value: 'add_inward',
        CommonPorts.OUTWARDS.value: 'add_outward',
        CommonPorts.MODEVARS_IN.value: 'add_inward_modevar',
        CommonPorts.MODEVARS_OUT.value: 'add_outward_modevar',
    }
    def log_debug(parent_attr, child_attr) -> str:
        logger.debug(
            f"{parent_attr} has been duplicated from {child_attr}"
            f" - including validation range and scope."
        )
    def copy_variable(
        child: System,
        port_name: str,
        child_var: str,
        parent_var: str,
        value: Any,
    ) -> None:
        """Copy variable `child_var` from `child` into `parent`, as `parent_var`."""
        child_port: BasePort = child[port_name]
        parent_port: BasePort = parent[port_name]
        lock_status = parent._locked
        parent._locked = False
        # Call `add_inward`, `add_outward`, etc., depending on context
        add_to_parent = getattr(parent, method_map[port_name])
        add_to_parent(parent_var, value)
        parent._locked = lock_status
        # Copy full variable detail in parent port
        parent_port.copy_variable_from(child_port, child_var, parent_var)
        log_debug(
            f"{parent.name}.{parent_var}",
            f"{parent.name}.{child.name}.{child_var}",
        )

    # Transform pulling into a name mapping (dict)
    if isinstance(pulling, str):
        pulling = {pulling: pulling}
    elif isinstance(pulling, dict):
        pass
    elif isinstance(pulling, Collection):
        pulling = dict(zip(pulling, pulling))

    for child_attr_name, parent_attr_name in pulling.items():
        child_attr = child[child_attr_name]

        if isinstance(child_attr, Port):
            if parent_attr_name not in parent:
                pulled_port = child_attr.copy(parent_attr_name)
                parent._add_port(pulled_port)
                log_debug(
                    f"Port {pulled_port.contextual_name}",
                    f"{parent.name}.{child_attr.contextual_name}",
                )
            else:
                pulled_port = parent[parent_attr_name]
            
            parent.connect(child_attr, pulled_port)

        else:  # ExtensiblePort (inwards or outwards)
            if isinstance(child_attr, BasePort):  # Pulling all inwards or outwards
                port_name = child_attr.name
                parent_port = parent[port_name]

                for varname, value in child_attr.items():
                    if varname not in parent_port:
                        copy_variable(child, port_name, varname, varname, value)

                parent.connect(child_attr, parent_port, list(child_attr))

            else:  # Pulling individual variable
                var = child.name2variable[child_attr_name]
                port_name = var.mapping.name

                if parent_attr_name not in parent[port_name]:
                    copy_variable(child, port_name, child_attr_name, parent_attr_name, child_attr)

                parent.connect(child[port_name], parent[port_name], {child_attr_name: parent_attr_name})



"""Various integration tests with multimode systems
"""
import pytest
import numpy

from cosapp.ports import Port
from cosapp.systems import System
from cosapp.drivers import EulerExplicit, NonLinearSolver
from cosapp.recorders import DataFrameRecorder
from typing import List


class ElecPort(Port):
    def setup(self):
        self.add_variable('V', 0.0, unit='V')
        self.add_variable('I', 1.0, unit='A')


class Resistor(System):
    """Resistor component

    Attributes:
        R : float
            Resistance in Ohms
    """
    def setup(self, R=0):
        self.add_input(ElecPort, 'elec_in')
        self.add_output(ElecPort, 'elec_out')

        self.add_inward('R', abs(float(R)), unit='ohm', desc='Internal resistance')
        self.add_outward('deltaV')

    def compute(self):
        self.elec_out.I = self.elec_in.I
        self.elec_out.V = self.elec_in.V - (self.elec_out.I * self.R)
        self.deltaV = self.elec_in.V - self.elec_out.V


class Node(System):
    """System representing an electric circuit node with
    an arbitrary number of incoming and outgoing branches.

    Nodes provide an off-design problem ensuring potential equality
    and global current balance (see 'Unknowns' and 'Equations' below).

    Constructor arguments:
    ----------------------
    - n_in [int], optional: Number of incoming branches. Defaults to 1.
    - n_out [int], optional: Number of outgoing branches. Defaults to 1.

    Properties:
    -----------
    - n_in [int]: Number of incoming branches.
    - n_out [int]: Number of outgoing branches.
    - incoming: Tuple containing all `ElecPort` inputs.
    - outgoing: Tuple containing all `ElecPort` outputs.

    Unknowns:
    ---------
    - n_out current fractions (one per outgoing branch), if n_out > 1.

    Equations:
    ----------
    - (n_in - 1) potential equality conditions for incoming branches.
    - 1 total current balance equation, if n_out > 1.
    """
    def setup(self, n_in=1, n_out=1):
        """Node constructor.

        Arguments:
        -----------
        - n_in [int], optional: Number of incoming branches. Defaults to 1.
        - n_out [int], optional: Number of outgoing branches. Defaults to 1.
        """
        self.add_property('n_in', int(n_in))
        self.add_property('n_out', int(n_out))

        if min(self.n_in, self.n_out) < 1:
            raise ValueError("Node needs at least one incoming and one outgoing branch")

        self.add_property('incoming',
            tuple(
                self.add_input(ElecPort, f"elec_in{i}")
                for i in range(self.n_in)
            )
        )
        self.add_property('outgoing',
            tuple(
                self.add_output(ElecPort, f"elec_out{i}")
                for i in range(self.n_out)
            )
        )

        if self.n_out > 1:  # unnecessary otherwise
            self.add_inward('I_frac',
                value = numpy.full(self.n_out, 1.0 / self.n_out),
                desc = f"Current fractions distributed to outgoing branches",
                limits = (0, 1),
            )
            self.add_unknown('I_frac', lower_bound=0, upper_bound=1)
            self.add_equation('sum(I_frac) == 1', name='Current balance')

        for i in range(1, self.n_in):   # case where node is 'joiner'
            self.add_equation(f'elec_in{i}.V == elec_in0.V')

        self.add_outward('V', 0., unit='V', desc='Actual node voltage')
        self.add_outward('sum_I_in', 0., unit='A', desc='Sum of all incoming currents')
        self.add_outward('sum_I_out', 0., unit='A', desc='Sum of all outgoing currents')

    def compute(self):
        # Sum of incoming currents
        self.sum_I_in = I = sum(port.I for port in self.incoming)

        # Output voltage
        self.V = V = numpy.mean([port.V for port in self.incoming])

        # Current distribution
        try:
            I_frac = self.I_frac
        except AttributeError:
            I_frac = [1]
        for j, port in enumerate(self.outgoing):
            port.V = V
            port.I = I * I_frac[j]

        self.sum_I_out = I * sum(I_frac)

    @classmethod
    def make(cls, parent, name, incoming: List[ElecPort], outgoing: List[ElecPort], pulling=None) -> "Node":
        """Factory method making appropriate connections with parent system"""
        node = cls(name, n_in=max(len(incoming), 1), n_out=max(len(outgoing), 1))
        parent.add_child(node, pulling=pulling)
        
        for branch_elec, node_elec in zip(incoming, node.incoming):
            parent.connect(branch_elec, node_elec)
        
        for branch_elec, node_elec in zip(outgoing, node.outgoing):
            parent.connect(branch_elec, node_elec)

        return node


class TwoWayCircuit(System):
    def setup(self):
        self.add_outward_modevar('upbranch', True)
        self.reconfig()

        self.add_outward('Requiv', 0.0)
        self.add_outward('deltaV', 0.0)
        self.add_event('switch', trigger='elec_in.V == elec_out.V')
    
    def compute(self):
        self.deltaV = self.elec_in.V - self.elec_out.V
        I = self.elec_in.I
        self.Requiv = self.deltaV / I if abs(I) > 0 else numpy.nan

    def transition(self):
        if self.switch.present:
            self.upbranch = not self.upbranch
            self.reconfig()

    def reconfig(self):
        for name in self.children:
            self.pop_child(name)
        pulled_ports = ['elec_in', 'elec_out']
        if self.upbranch:
            child = self.add_child(Resistor("R1", R=100), pulling=pulled_ports)
        else:
            child = self.add_child(Resistor("R2", R=500), pulling=pulled_ports)
        self.elec_out.V = child.elec_out.V = 0.0


class TwoWayCircuitWithEq(System):
    def setup(self):
        self.add_outward_modevar('upbranch', True)
        self.add_child(Resistor("R0", R=100), pulling=['elec_in', 'elec_out'])
        self.reconfig()

        self.add_outward('Requiv', 0.0)
        self.add_outward('deltaV', 0.0)
        self.add_event('switch', trigger='elec_in.V == elec_out.V')
    
    def compute(self):
        self.deltaV = self.elec_in.V - self.elec_out.V
        I = self.elec_in.I
        self.Requiv = self.deltaV / I if abs(I) > 0 else numpy.nan

    def transition(self):
        if self.switch.present:
            self.upbranch = not self.upbranch
            self.reconfig()

    def reconfig(self):
        problem = self._math
        if not self.upbranch and problem.is_empty():
            self.add_unknown("R0.R").add_equation("R0.R == 500")
        else:
            problem.clear()
            self.R0.R = 100


@pytest.fixture
def case_TwoWayCircuit():
    circuit = TwoWayCircuit('circuit')
    driver = circuit.add_driver(
        EulerExplicit(time_interval=[0, 1], dt=0.1)
    )
    driver.add_recorder(
        DataFrameRecorder(includes=['elec_in.*', 'elec_out.V', 'Requiv', 'deltaV', 'upbranch']),
        period = 0.1,
    )
    solver = driver.add_child(NonLinearSolver('solver', tol=1e-9))
    solver.add_unknown('elec_in.I').add_equation("elec_out.V == 0")
    return circuit, driver


def test_TwoWayCircuit(case_TwoWayCircuit):
    circuit, driver = case_TwoWayCircuit
    omega = 6
    driver.set_scenario(
        values = {
            'elec_in.V': f"cos({omega} * t)",
        }
    )
    circuit.run_drivers()

    df = driver.recorder.export_data()
    # print("", df.drop(['Section', 'Status', 'Error code'], axis=1), sep="\n")

    for i, row in df.iterrows():
        I = row['elec_in.I']
        R = row['Requiv']
        context = f"row #{i}, I = {I}, {list(circuit.exec_order)}"
        if I > 1e-12:
            assert R == pytest.approx(100), context
        elif I < -1e-12:
            assert R == pytest.approx(500), context
    
    assert [record.time for record in driver.recorded_events] == pytest.approx(
        [(2 * k + 1) * 0.5 * numpy.pi / omega for k in range(2)]
    )


@pytest.fixture
def case_TwoWayCircuitWithEq():
    circuit = TwoWayCircuitWithEq('circuit')
    driver = circuit.add_driver(
        EulerExplicit(time_interval=[0, 1], dt=0.1)
    )
    driver.add_recorder(
        DataFrameRecorder(includes=['elec_in.*', 'elec_out.V', 'Requiv', 'deltaV', 'upbranch']),
        period = 0.1,
    )
    solver = driver.add_child(NonLinearSolver('solver', tol=1e-9))
    solver.add_unknown('elec_in.I').add_equation("elec_out.V == 0")
    return circuit, driver


def test_TwoWayCircuitWithEq(case_TwoWayCircuitWithEq):
    circuit, driver = case_TwoWayCircuitWithEq
    omega = 6
    driver.set_scenario(
        values = {
            'elec_in.V': f"cos({omega} * t)",
        }
    )
    circuit.run_drivers()

    df = driver.recorder.export_data()
    # print("", df.drop(['Section', 'Status', 'Error code'], axis=1), sep="\n")

    for i, row in df.iterrows():
        I = row['elec_in.I']
        R = row['Requiv']
        context = f"row #{i}, I = {I}"
        if I > 1e-12:
            assert R == pytest.approx(100), context
        elif I < -1e-12:
            assert R == pytest.approx(500), context
    
    assert [record.time for record in driver.recorded_events] == pytest.approx(
        [(2 * k + 1) * 0.5 * numpy.pi / omega for k in range(2)]
    )



"""
Event handling and discrete stepping
"""
import scipy.optimize
from math import inf

from typing import NamedTuple, Iterator, List, Dict, Tuple
from cosapp.multimode.event import Event
from cosapp.utils import partition


class TimedEvent(NamedTuple):
    """Named tuple associating an event and its occurrence time"""
    event: Event
    time: float

    @classmethod
    def empty(cls):
        return cls(None, inf)


class DiscreteStepper():

    __slots__ = (
        '_owner', '_state', '_system', '_sysview',
        '_interval', '_nonprimitives', '_primitives',
    )

    def __init__(self, driver: "Driver"):
        # Local import to avoid cyclic dependency
        from cosapp.drivers.time.utils import SystemInterpolator

        self._owner = driver
        self._system = driver.owner
        self._sysview = SystemInterpolator(driver)
        self._interval = None
        self.set_events()

    def set_events(self) -> None:
        """Update event list from system of interest,
        and from stop criterion of owner time driver.
        """
        self._state: Dict[Event, bool] = dict.fromkeys(self._system.all_events(), False)
        stop = self._owner.scenario.stop
        if stop is not None:
            self._state[stop] = False
        primitives, non_primitives = partition(self.events(), lambda event: event.is_primitive)
        self._primitives: List[Event] = primitives
        self._nonprimitives: List[Event] = non_primitives

    def reset(self) -> None:
        """Update event list, and reset all events"""
        self.set_events()
        for event in self.events():
            event.reset()

    def events(self) -> Iterator[Event]:
        """Iterator on handled events"""
        return self._state.keys()

    def present_events(self) -> Iterator[Event]:
        """Returns an iterator on all present events"""
        return filter(lambda event: event.present, self.events())

    @property
    def sysview(self) -> "SystemInterpolator":
        """SystemInterpolator: system interpolator"""
        return self._sysview

    @property
    def interval(self) -> Tuple[float, float]:
        """Tuple[float, float]: time interval"""
        return self._interval

    @interval.setter
    def interval(self, interval: Tuple[float, float]):
        self._interval = interval

    def set_data(self, interval, interpol) -> None:
        """Sets interpolation data used for detecting the first event that occurred in a given time step."""
        self.interval = interval
        self._sysview.interp = interpol

    def trigger_time(self, event: Event) -> float:
        """Returns the date at which a primitive event was triggered.
        This method may only be called once the list of all triggered events is known
        and the interpolation data has been set."""
        sysview = self._sysview
        def f(t):
            sysview.exec(t)
            return event.value()
        t1, t2 = self._interval
        t_event = scipy.optimize.brentq(f, t1, t2)
        # Reset system at original state
        sysview.exec(t1)
        return t_event

    def find_primal_event(self) -> TimedEvent:
        """Returns a `TimedEvent` named tuple containing the first primitive event triggered,
        together with its occurrence date ((None, inf) if no primitive event is triggered).
        The internal state is only updated if an event is triggered.
        This method should only be called at the first microstep of the first discrete time step.
        """
        # TODO: Possibility of several events at the same time (up to rounding errors)?
        # For instance, if an event occurs every 0.2s and another event every s...
        triggered_events = list(filter(lambda e: e.present, self._primitives))
        primal = TimedEvent.empty()
        for event in triggered_events:
            time = self.trigger_time(event)
            if time < primal.time:
                primal = TimedEvent(event, time)
        self._state[primal.event] = True
        self._sysview.exec(primal.time)
        return primal

    def event_detected(self) -> bool:
        """Tests all primitive events and returns a Boolean value indicating
        whether at least one of these events is to be triggered."""
        return any(event.to_trigger() for event in self._primitives)

    def __microstep(self) -> bool:
        """Performs an evaluation microstep for non-primitive events and returns a Boolean
        indicating whether one or several events were triggered in the current cascade.
        This method should not be called at the first microstep of a discrete time step.
        """
        has_changed = False
        for event in self._nonprimitives:
            triggered = event.step()
            self._state[event] |= triggered
            has_changed |= triggered
        return has_changed

    def first_discrete_step(self) -> TimedEvent:
        """Performs the first discrete step and returns a `TimedEvent` indicating
        which primitive event was triggered and at which date."""
        for event in self._primitives:
            event.step()
        occur = self.find_primal_event()
        while self.__microstep():
            pass
        return occur

    def discrete_step(self) -> List[Event]:
        """Performs a discrete step other than the first one"""
        for event in self._primitives:
            self._state[event] = event.step()
        while self.__microstep():
            pass
        return list(self.present_events())

    def tick(self):
        """Ends a discrete loop by ticking all events."""
        for event in self.events():
            event.tick()

    def shift(self) -> None:
        """Performs a time shift of all primitive events."""
        for event in self._primitives:
            event.tick()

    def reevaluate_primitive_events(self) -> None:
        for event in self._primitives:
            event.reevaluate()


"""
`Driver`s for `System` optimization calculation.
"""
import numpy
import scipy.optimize
import warnings
import copy
from numbers import Number
from typing import (
    Union, Iterable, Callable, Optional,
    Any, Sequence, Dict, List, Set, Tuple,
)
from collections.abc import Collection

from cosapp.core.eval_str import EvalString
from cosapp.core.numerics.basics import MathematicalProblem, SolverResults
from cosapp.core.numerics.boundary import Unknown
from cosapp.drivers.abstractsolver import AbstractSolver
from cosapp.drivers.optionaldriver import OptionalDriver
from cosapp.drivers.utils import ConstraintParser, dealias_problem
from cosapp.recorders.recorder import BaseRecorder
from cosapp.utils.options_dictionary import OptionsDictionary
from cosapp.utils.helpers import check_arg

import logging
logger = logging.getLogger(__name__)


# TODO
# [ ] Pull unknowns from Systems
class Optimizer(AbstractSolver):
    """Driver running an optimization problem on its `System` owner.

    In general, the optimization problems are of the form::

        minimize f(x) subject to

        g_i(x) >= 0,  i = 1,...,m
        h_j(x)  = 0,  j = 1,...,p

    where ``x`` is a vector of one or more variables. ``g_i(x)`` are the inequality constraints.
    ``h_j(x)`` are the equality constrains.

    Optionally, the lower and upper bounds for each element in x can also be specified.

    Parameters
    ----------
    name : str
        Name of the driver
    owner : System, optional
        :py:class:`~cosapp.systems.system.System` to which driver belongs; defaults to `None`
    **kwargs : Any
        Keyword arguments will be used to set driver options

    Attributes
    ----------
    name : str
        Name of the driver
    parent : Driver, optional
        Top driver containing this driver; default None.
    owner : System
        :py:class:`~cosapp.systems.system.System` to which this driver belong.
    children : OrderedDict[Driver]
        Drivers belonging to this one.

    options : OptionsDictionary
      |  Options for the current driver
      |  **verbose** : int, {0, 1}
      |    Verbosity level of the driver; default 0 (i.e. minimal information)
      |  **eps** : float, [1.5e-8, 1.]
      |    Step size used for numerical approximation of the Jacobian; default 1.5e-8
      |  **ftol** : float, [1.5e-8, 1.]
      |    Iteration termination criteria (f^k - f^{k+1})/max{\|f^k\|,\|f^{k+1}\|,1} <= ftol;
      |    default 1e-6
      |  **max_iter** : int, [1, [
      |    Maximum number of iterations; default 100

    solution :  # TODO

    Notes
    -----
    This optimizer is a wrapper around ``scipy.optimize.minimize`` function. For more details please
    refer to:
    https://docs.scipy.org/doc/scipy-1.0.0/reference/generated/scipy.optimize.minimize.html

    """
    __slots__ = ('_constraints', '_raw_constraints', '_initial_state', '_objective', '__current_x')

    def __init__(self,
        name: str,
        owner: Optional["cosapp.systems.System"] = None,
        **kwargs
    ) -> None:
        """Initialize driver

        Parameters
        ----------
        name: str, optional
            Name of the `Driver`.
        owner: System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; defaults to `None`.
        **kwargs:
            Additional keywords arguments forwarded to base class.
        """
        super().__init__(name, owner, **kwargs)

        # TODO we need to move this in an enhanced MathematicalProblem
        self._raw_constraints: Set[str] = set()  # Human-readable constraints
        self._constraints: List[Dict] = list()   # Non-negativity constraints
        self._initial_state: Dict[str, Any] = dict()
        self._objective: EvalString = None
        self.__current_x = numpy.empty(0)

        self.options.declare(
            name = 'method',
            default = None,
            values = self.available_methods(),
            desc = (
                "Type of solver."
                " If not given, chosen to be one of 'BFGS', 'L-BFGS-B', 'SLSQP',"
                " depending if the problem has constraints or bounds."
            ),
            allow_none = True,
        )
        self.options.declare(
            'eps', 2**(-26), dtype=float, lower=2**(-26), upper=1.,
            desc="Step size used for numerical approximation of the Jacobian.",
        )
        self.options.declare(
            'ftol', 1.5e-8, dtype=float, lower=1e-15, upper=1.,
            desc="Iterations stop when (f^k - f^{k+1}) / max(|f^k|, |f^{k+1}|, 1) <= ftol.",
        )
        self.options.declare(
            'maxiter', 100, dtype=int, lower=1,
            desc='Maximum number of iterations.',
        )
        self.options.declare(
            'monitor', False, dtype=bool, allow_none=False,
            desc="Defines if intermediate system state should be recorded.",
        )
        self._filter_options(kwargs, aliases={'tol': 'ftol', 'max_iter': 'maxiter'})

    def set_minimum(self, expression: str) -> None:
        """Set the scalar objective function to be minimized.

        Parameters
        ----------
        expression : str
            The objective expression to be minimized.
        """
        self.__check_owner("objective")
        check_arg(expression, "expression", str, lambda s: "==" not in s)

        self._objective = objective = EvalString(expression, self.owner)
        if objective.constant:
            warnings.warn(f"Objective is constant {objective.eval()}")

    def set_maximum(self, expression: str) -> None:
        """Set the scalar objective function to be maximized.

        Parameters
        ----------
        expression : str
            The objective expression to be maximized.
        """
        self.set_minimum(f"-({expression})")

    def set_objective(self, expression: str) -> None:
        """Set the scalar quantity to be minimized.
        Same as `set_minimum`.

        Note:
        -----
        This method is deprecated, and should be replaced by either
        `set_minimum` or `set_maximum`, depending on the objective.

        Parameters
        ----------
        expression : str
            The objective expression to be minimized.
        """
        warnings.warn(
            "method `set_objective` is deprecated; use `set_minimum` or `set_maximum` instead."
        )
        self.set_minimum(expression)

    @staticmethod
    def available_methods() -> List[str]:
        """Returns all possible values of option `method`.
        For more information, please refer to the online documentation
        of `scipy.optimize.minimize`.
        """
        return [
            'Nelder-Mead', 'Powell', 'CG', 'BFGS', 'Newton-CG',
            'L-BFGS-B', 'TNC', 'COBYLA', 'SLSQP', 'dogleg',
            'trust-constr', 'trust-ncg', 'trust-exact', 'trust-krylov',
        ]

    @property
    def objective(self):
        try:
            return self._objective.eval()
        except:
            return None

    @property
    def objective_expr(self) -> str:
        return str(self._objective) if self._objective else None

    def add_unknown(self,
            name: Union[str, Iterable[Union[dict, str, Unknown]]],
            max_abs_step: Number = numpy.inf,
            max_rel_step: Number = numpy.inf,
            lower_bound: Number = -numpy.inf,
            upper_bound: Number = numpy.inf
    ) -> "MathematicalProblem":
        """Add unknown variables.

        You can set variable one by one or provide a list of dictionary to set multiple variable at once. The
        dictionary key are the arguments of this method.

        Parameters
        ----------
        name : str or Iterable of dictionary or str
            Name of the variable or list of variable to add
        max_rel_step : float, optional
            Maximal relative step by which the variable can be modified by the numerical solver; default numpy.inf
        max_abs_step : float, optional
            Maximal absolute step by which the variable can be modified by the numerical solver; default numpy.inf
        lower_bound : float, optional
            Lower bound on which the solver solution is saturated; default -numpy.inf
        upper_bound : float, optional
            Upper bound on which the solver solution is saturated; default numpy.inf

        Returns
        -------
        MathematicalProblem
            The modified MathematicalSystem
        """
        self.__check_owner("variables")
        self._raw_problem.add_unknown(name, max_abs_step, max_rel_step, lower_bound, upper_bound)

    def add_constraints(self, expression: Union[str, List[str]]) -> None:
        """Add constraints to the optimization problem.

        Parameters
        ----------
        - expression [str or List[str]]:
            Human-readable equality or inequality constraints, such as
            'x >= y**2', '0 < alpha < 1', 'a == b', or a list thereof.
        
        Note
        ----
        Expressions are parsed into non-negative constraints in the
        optimization problem. Strict inequalities are not enforced,
        and treated as non-strict inequalities.
        For instance, `x < y` translates into: `y - x >= 0`.
        """
        self.__check_owner("constraints")
        check_arg(expression, 'expression', (str, Collection))

        if isinstance(expression, str):
            expression = [expression] 
        
        self._raw_constraints.update(expression)
        constraints = ConstraintParser.parse(expression)

        for constraint in constraints:
            # Test that expression can be evaluated
            try:
                evalstr = EvalString(constraint.expression, self.owner)
            except:
                raise
            data = dict(
                type = 'ineq' if constraint.is_inequality else 'eq',
                expr = evalstr,
            )
            self._constraints.append(data)

    @property
    def constraints(self) -> Set[str]:
        """Set[str]: representation of optimization constraints."""
        return self._raw_constraints.copy()

    def __check_owner(self, kind: str) -> None:
        if self.owner is None:
            raise AttributeError(f"Owner system is required to define optimization {kind}.")

    def setup_run(self):
        """Method called once before starting any simulation."""
        super().setup_run()
        self.__current_x = numpy.empty(0)
        
        # Resolve unknown aliasing and connected unknowns
        self.problem = dealias_problem(self._raw_problem)
        self.touch_unknowns()

    def _expression_wrapper(self, expression: EvalString) -> Callable[[numpy.ndarray], float]:
        """Wrapper around objective and constraint expression to propagate the x values in the
            `System` owner.

        Parameters
        ----------
        expression : EvalString
            The expression to be evaluated.

        Returns
        -------
        Callable[[numpy.ndarray], float]
            Callable objective function usable by scipy.optimize.minimize
        """
        if self.children:
            def wrapper(x: numpy.ndarray, *args) -> float:
                modified = self._update_unknowns(x)
                if modified:
                    for driver in self.children.values():
                        driver.run_once()
                return expression.eval()

        else:
            def wrapper(x: numpy.ndarray, *args) -> float:
                modified = self._update_unknowns(x)
                if modified:
                    self.owner.run_children_drivers()
                return expression.eval()

        return wrapper

    def _fresidues(self, x: numpy.ndarray) -> float:
        """
        Method used by the solver to take free variables values as input and values of the objective function (after
        running the System).

        Parameters
        ----------
        x : numpy.ndarray
            The list of values to set to the free variables of the `System`

        Returns
        -------
        float
            Objective function value
        """
        x = numpy.asarray(x)
        logger.debug(f"Call fresidues with x = {x!r}")
        self.set_iteratives(x)

        if self.children:
            for subdriver in self.children.values():
                subdriver.run_once()
        else:
            self.owner.run_children_drivers()

        objective = self._objective.eval()
        logger.debug(f"Objective: {objective!r}")
        return objective

    def set_iteratives(self, x: Sequence[float]) -> None:
        self._update_unknowns(x)

    def _update_unknowns(self, x: Sequence[float]) -> bool:
        modified = not numpy.array_equal(x, self.__current_x)

        if modified:
            x = numpy.asarray(x)
            counter = 0
            for unknown in self.problem.unknowns.values():
                if unknown.mask is None:
                    unknown.set_default_value(x[counter])
                    counter += 1
                else:
                    n = numpy.count_nonzero(unknown.mask)
                    unknown.set_default_value(x[counter : counter + n])
                    counter += n
                # Set variable to new x
                unknown.set_to_default()

            self.__current_x = numpy.array(x)  # force copy

        return modified

    def resolution_method(self,
        fresidues: Callable[[Sequence[float]], float],
        x0: numpy.ndarray,
        args: Tuple[Union[float, str], bool] = (),
        options: Optional[OptionsDictionary] = None,
        bounds = None,
        constraints = None,
    ) -> SolverResults:
        """Function call to cancel the residues.

        Parameters
        ----------
        fresidues : Callable[[Sequence[float], Union[float, str]], float]
            Residues function taking two parameters (evaluation vector, time/ref) and returning the residues
        x0 : numpy.ndarray
            The initial values vector to converge to the solution
        args : Tuple[Union[float, str], bool], optional
            A tuple of additional argument for fresidues starting with the time/ref parameter
        options : OptionsDictionary, optional
            Options for the numerical resolution method

        Returns
        -------
        SolverResults
            Solution container
        """
        sub_options = {
            # Specify both `ftol` and `gtol`, as name varies among scipy solvers
            'ftol': options['ftol'],
            'gtol': options['ftol'],  # `ftol` is not understood by unconstrained solvers
            'eps': options['eps'],
            'maxiter': options['maxiter'],
            'disp': bool(options['verbose']),
        }

        def callback(*args, **kwargs) -> bool:
            self._record_data()
            return False

        with warnings.catch_warnings():
            # Ignore warnings about `gtol` or `ftol` potentially emitted by scipy solver
            warnings.filterwarnings("ignore", message="Unknown solver options: [fg]tol")
            output = scipy.optimize.minimize(
                fresidues, x0,
                args=args,
                tol=options['ftol'],
                method=options['method'],
                bounds=bounds,
                constraints=constraints,
                options=sub_options,
                callback=callback,
            )
        if output.fun < self.objective:
            # Solver solution does not correspond to
            # last system execution; rerun to synch.
            self._fresidues(output.x)
            self._record_data()
        return output

    def _precompute(self):
        """List unknowns and gather initial values."""
        super()._precompute()
        OptionalDriver.set_inhibited(True)

        init = numpy.empty(0)

        for name, unknown in self.problem.unknowns.items():
            if not self.force_init and name in self.solution:
                # We ran successfully at least once and are environmental friendly
                data = self.solution[name]

            else:  # User wants the init or first simulation or crash
                try:
                    boundary = self._initial_state[name]
                except KeyError:
                    data = copy.deepcopy(unknown.value)
                else:
                    umask = unknown.mask if unknown.mask is not None else numpy.empty(0)
                    bmask = boundary.mask if boundary.mask is not None else numpy.empty(0)
                    if not numpy.array_equal(umask, bmask):
                        raise ValueError(
                            f"Unknown and initial conditions on {unknown.name!r} are not masked equally"
                        )
                    data = copy.deepcopy(boundary.default_value)
                    # self._initial_state[name] = boundary

            init = numpy.append(init, data)

        self.initial_values = init

    def compute(self) -> None:
        """Execute the optimization."""
        self.status = ''
        self.error_code = '0'

        # Check that objective is set
        if self._objective is None:
            self.status = 'ERROR'
            self.error_code = '9'
            raise ArithmeticError(
                "Optimization objective was not specified."
            )

        # Gather and simplify the bounds
        limits = self._get_solver_limits()
        unique_lower = numpy.unique(limits['lower_bound'])
        unique_upper = numpy.unique(limits['upper_bound'])
        def get_bounds(lower, upper) -> tuple:
            return (
                None if lower == -numpy.inf else lower,
                None if upper ==  numpy.inf else upper
            )
        if unique_lower.size == unique_upper.size == 1:
            bounds = get_bounds(unique_lower[0], unique_upper[0])
            if bounds == (None, None):
                bounds = None
        else:
            bounds = tuple()

        if bounds is not None:
            bounds = [
                get_bounds(lower, upper)
                for lower, upper in zip(limits['lower_bound'], limits['upper_bound'])
            ]

        # Create nonlinear constraints
        # TODO The following is really ugly. Constraints should be merged
        # from children through MathematicalProblem extension.
        def format_constraint(constraint):
            return {
                'type': constraint['type'],
                'fun': self._expression_wrapper(constraint['expr'])
            }
        constraints = tuple(map(format_constraint, self._constraints))

        if len(self.initial_values) > 0:
            monitored = self.options['monitor']
            BaseRecorder.paused = not monitored
            if monitored:
                self._fresidues(self.initial_values)
                self._record_data()

            results = self.resolution_method(
                self._fresidues,
                self.initial_values,
                args = (),
                bounds = bounds,
                constraints = constraints,
                options = self.options,
            )

            if not results.success:
                self.status = 'ERROR'
                self.error_code = '9'
                self.solution = {}
                logger.error(f"The solver failed: {results.message}")
            else:
                self.solution = dict(
                    (key, unknown.default_value)
                    for key, unknown in self.problem.unknowns.items()
                )
                self._print_solution()

            if not monitored:
                BaseRecorder.paused = False
                self._record_data()

        else:
            logger.warning('No design variable has been specified for the optimization.')

    def _record_data(self) -> None:
        """Record data into recorder, if any."""
        if self._recorder is not None:
            self._recorder.record_state(self.name)

    def _postcompute(self) -> None:
        """Undo pull inputs and reset iteratives sets."""
        OptionalDriver.set_inhibited(False)
        super()._postcompute()

    def _print_solution(self) -> None:  # TODO better returning a string
        """Print the solution in the log."""
        if self.options['verbose']:
            logger.info(f"Objective function: {self._objective.eval():.5g}")
            logger.info(f"Parameters [{len(self.solution)}]: ")
            for name, value in self.solution.items():
                logger.info(f"   # {name}: {value}")
            constraints = self._constraints
            if constraints:
                logger.info(f"Constraints [{len(constraints)}]: ")
                for constraint in constraints:
                    expr = constraint['expr']
                    logger.info(f"   # {expr}: {expr.eval()}")

    def _repr_markdown_(self) -> str:
        """Mardown representation of optimization problem."""
        itemize = lambda item: f"* {item}"

        def section(header: str, collection: Union[str, List[str]]) -> List[str]:
            content = []
            if isinstance(collection, str):
                collection = [collection]
            if collection:
                content.append(f"#### {header.title()}:\n")
                content.extend(map(itemize, collection))
                content.append("")
            return content

        lines = []

        def add_section(header, collection) -> None:
            nonlocal lines
            lines.extend(section(header, collection))
        
        add_section("Objective", f"Minimize {self._objective}")
        add_section("Unknowns", self._raw_problem.unknowns)
        add_section("Constraints", self._raw_constraints)

        return "\n".join(lines)



import pytest
import numpy as np

from cosapp.drivers.time.utils import TwoPointCubicPolynomial


@pytest.mark.parametrize("xs", [[1, 2], [0, 0.5], [-0.2, 0.1]])
@pytest.mark.parametrize("ys", [[0.7, -0.2], [0, 1]])
@pytest.mark.parametrize("dy", [[-0.5, 0.1], [0, 1], [0, 0]])
def test_TwoPointCubicPolynomial(xs, ys, dy):
    xs = np.asarray(xs)
    ys = np.asarray(ys)
    dy = np.asarray(dy)
    p = TwoPointCubicPolynomial(xs, ys, dy)
    dp = p.deriv()

    assert p.degree() == 3
    assert p(xs) == pytest.approx(ys, abs=1e-14)
    assert dp(xs) == pytest.approx(dy, abs=1e-14)


def test_numpy_polynomial():
    p = np.polynomial.Polynomial([0, 1], [1, 2], [0, 1])

    assert p(0) == pytest.approx(-1, abs=1e-14)
    assert p(2) == pytest.approx(+1, abs=1e-14)
